#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 17 03:28:27 2019

@author: carl
"""

import os
import sys
from pyshortcuts import make_shortcut
import argparse

def main():
    parser = argparse.ArgumentParser(
            description='Creates icons for the OCTAVVS scripts.')
    _ = parser.parse_args()

    executables = {'oct_preprocessing': 'Preprocessing',
#                   'oct_mcr_als': 'MCR-ALS',
#                   'oct_clustering': 'Clustering',}
                   'oct_decomposition': 'Decomposition'}
    proj = 'OCTAVVS '

    for cmd, nom in executables.items():
        icon = os.path.join(os.path.split(__file__)[0], 'prep', 'octavvs_prep')
        script = os.path.join(os.path.split(sys.argv[0])[0], cmd)

        make_shortcut(script, name=proj+nom, icon=icon,
                      startmenu=False, terminal=True, desktop=True)

if __name__ == '__main__':
    main()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 21:07:55 2020

@author: carl
"""

import os
import sys
import traceback
import multiprocessing

from PyQt5.QtWidgets import QMainWindow, QApplication, QFileDialog
from PyQt5.QtWidgets import QErrorMessage, QMessageBox #, QInputDialog, QDialog
from PyQt5.QtCore import Qt, QSettings
from PyQt5.Qt import qApp

from .exceptiondialog import ExceptionDialog
from .copyfigure import add_clipboard_to_figures
from .. import octavvs_version



class OctavvsMainWindow(QMainWindow):

    fileOptions = QFileDialog.Options() | QFileDialog.DontUseNativeDialog

    def __init__(self, parent=None):
        super().__init__(parent)
        self.settings = QSettings('MICCS', 'OCTAVVS ' + self.program_name())
        qApp.installEventFilter(self)
        self.setupUi(self)

        self.errorMsg = QErrorMessage(self)
        self.errorMsg.setWindowModality(Qt.WindowModal)

        self.units = 'cm-1'
        self.plot_ltr = True

    def post_setup(self):
        "Called after setting up UI but before loading data etc"
        self.setWindowTitle('OCTAVVS %s %s' % (self.program_name(), octavvs_version))
        ExceptionDialog.install(self)

    @classmethod
    def program_name(cls):
        "Return the name of the program that this main window represents"
        return cls.__name__

    def showDetailedErrorMessage(self, err, details):
        "Show an error dialog with details"
        q = QMessageBox(self)
        q.setIcon(QMessageBox.Critical)
        q.setWindowTitle("Error")
        q.setText(err)
        q.setTextFormat(Qt.PlainText)
        q.setDetailedText(details)
        q.addButton('OK', QMessageBox.AcceptRole)
        return q.exec()

    def loadErrorBox(self, file, err, warning=False):
        "Prepare an error dialog for failure to load a file"
        q = QMessageBox(self)
        if warning:
            q.setIcon(QMessageBox.Information)
            q.setWindowTitle("Warning loading file")
            q.setText("Warning from loading '"+file+"':\n"+err[0])
        else:
            q.setIcon(QMessageBox.Warning)
            q.setWindowTitle("Error loading file")
            q.setText("Failed to load '"+file+"':\n"+err[0])
        q.setTextFormat(Qt.PlainText)
        q.setDetailedText(err[1])
        return q

    def getLoadSaveFileName(self, title, filter=None, settingname=None,
                            savesuffix=None, multiple=False,
                            settingdefault=None,
                            directory=None, defaultfilename=None):
        "Show a file dialog and select one or more files"
        setting = self.settings.value(settingname, settingdefault
                                      ) if settingname is not None else None
        if directory is None:
            directory = setting if type(setting) is str else None
        dialog = QFileDialog(parent=self, caption=title,
                             directory=directory, filter=filter)
        if defaultfilename is not None:
            dialog.selectFile(defaultfilename)
#        if setting and type(setting) is not str:
#            dialog.restoreState(setting)
        dialog.setOption(QFileDialog.DontUseNativeDialog, True)
        if savesuffix is not None:
            dialog.setAcceptMode(QFileDialog.AcceptSave)
            dialog.setDefaultSuffix(savesuffix)
        elif multiple:
            dialog.setFileMode(QFileDialog.ExistingFiles)
        else:
            dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.exec()
        files = dialog.selectedFiles()
        if not dialog.result() or not files:
            return None
        if settingname is not None:
            self.settings.setValue(settingname, os.path.dirname(files[0]))
        return files if multiple else files[0]

    def getSaveFileName(self, title, suffix='', **kwargs):
        "Show a file dialog and select an output file"
        return self.getLoadSaveFileName(
            title=title, savesuffix=suffix, **kwargs)

    def getLoadFileName(self, title, **kwargs):
        "Show a file dialog and select an input file"
        return self.getLoadSaveFileName(title=title, **kwargs)

    def getLoadFileNames(self, title, **kwargs):
        "Show a file dialog and select an input file"
        return self.getLoadSaveFileName(title=title, multiple=True, **kwargs)

    def getImageFileName(self, title, **kwargs):
        "Show a file dialog and select a single image file"
        return self.getLoadSaveFileName(title=title,
                filter="Image files (*.jpg *.jpeg *.png *.tif "
                    "*.tiff *.bmp *.gif);;All files (*)",
                **kwargs)

    def getDirectoryName(self, title, settingname=None, savesetting=True,
                         default=None):
        "Show a file dialog and select a directory"
        directory = self.settings.value(settingname, default) if \
            settingname is not None else None
        dialog = QFileDialog(parent=self, caption=title,
                             directory=directory)

        dialog.setAcceptMode(QFileDialog.AcceptSave)
        dialog.setFileMode(QFileDialog.Directory)
        dialog.setOption(QFileDialog.DontUseNativeDialog, True)
        dialog.setOption(QFileDialog.ShowDirsOnly, True)
        dialog.exec()
        dirs = dialog.selectedFiles()
        if not dialog.result() or not dirs:
            return None
        if savesetting:
            self.settings.setValue(settingname, dirs[0])
        return dirs[0]

    def updateWavenumberRange(self):
        "For derived classes, when the wn range is updated"
        pass

    def updateDimensions(self, wh):
        "For derived classes, when image width/height is updated"
        pass

    def updateUnitsAndOrder(self, units, ltr):
        "Update the units (cm^-1 or nm) and direction of plots"
        self.units = units
        self.plot_ltr = ltr

    def unitsRichText(self):
        if self.units == 'cm-1':
            return 'cm<sup>-1</sup>'
        return self.units

    @classmethod
    def run_octavvs_application(windowclass, parser=None, parameters=[],
                                isChild=False):
        res = 1
        try:
            progver = 'OCTAVVS %s %s' % (windowclass.program_name(),
                                         octavvs_version)
            windowparams = {}
            if parser is not None:
                parser.add_argument('--version', action='version',
                                    version=progver)
                selmp = multiprocessing.get_start_method(allow_none=True) is None
                if selmp:
                    parser.add_argument('--mpmethod', help='')
                args = parser.parse_args()
                windowparams = { k: args.__dict__[k] for k in parameters }
                if selmp and args.mpmethod:
                    multiprocessing.set_start_method(args.mpmethod)

            app = QApplication.instance()
            if not app:
                app = QApplication(sys.argv)
            add_clipboard_to_figures()
            window = windowclass()
            window.show()
            window.post_setup(**windowparams)
            if not isChild:
                qApp.lastWindowClosed.connect(qApp.quit);
                res = app.exec_()

        except Exception:
            traceback.print_exc()
            print('Press some key to quit')
            input()
        if not isChild :
            sys.exit(res)




from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from matplotlib.figure import Figure
# from matplotlib.ticker import MaxNLocator
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt
#import matplotlib.patches as patches
import numpy as np
#import random



class DataPlotWidget(FigureCanvas):
    updated = pyqtSignal()
    clicked = pyqtSignal()

    def __init__(self, parent=None):
        self.fig = Figure(tight_layout=dict(pad=0.4))
        self.ax = self.fig.add_subplot(111)
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
#        self.fig.canvas.mpl_connect('button_press_event', self.onclick)
        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding,
                                   QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        self.wavenumber = None
        self.inspectra = None
        self.spectra = None
        self.popax = None
        self.ltr = True
        self.window_title = self.__class__.__name__

    def mousePressEvent(self, event):
        self.clicked.emit()

    def setData(self, wavenumber, inspectra, spectra):
        if (self.wavenumber is wavenumber and
            self.inspectra is inspectra and
            self.spectra is spectra):
            return
        # Skip redraw if there is and was no outdata
        if self.spectra is not None or spectra is not None:
            self.draw_idle()
        self.wavenumber = wavenumber
        self.inspectra = inspectra
        self.spectra = spectra
        self.updated.emit()

    def getSpectra(self):
        return self.spectra if self.spectra is not None else self.inspectra

    def getWavenumbers(self):
        return self.wavenumber

    def setOrder(self, ltr):
        if ltr != self.ltr:
            if ltr == self.ax.xaxis_inverted():
                self.ax.invert_xaxis()
                FigureCanvas.draw(self)
            if self.popax and ltr == self.popax.xaxis_inverted():
                self.popax.invert_xaxis()
                self.popax.figure.canvas.draw()
            self.ltr = ltr

    def draw_ax(self, ax):
        pass

    def draw_ax_single(self, ax, args={}, spectra=None):
        if spectra is None:
            spectra = self.spectra
        lines = ax.get_lines()
        if len(lines) == len(spectra):
            for i, l in enumerate(lines):
                l.set_data(self.wavenumber, spectra[i])
            ax.relim()
            ax.autoscale_view()
        else:
            ax.clear()
            for i in spectra:
                ax.plot(self.wavenumber, i, linewidth=1, **args)

    def draw_ax_double(self, ax, label1, label2, spectra1=None,
                       spectra2=None, args1={}, args2={}):
        if spectra1 is None:
            spectra1 = self.inspectra
        if spectra2 is None:
            spectra2 = self.spectra
        lines = ax.get_lines()
        nspec = len(self.spectra)
        if len(lines) == nspec * 2:
            for i, l in enumerate(lines):
                l.set_data(
                    self.wavenumber, spectra1[i] if i < nspec
                    else spectra2[i - nspec])
            ax.relim()
            ax.autoscale_view()

            if ax.figure.canvas.toolbar is not None:
                tb = ax.figure.canvas.toolbar
                if tb._nav_stack._elements:
                    wl = tb._nav_stack._elements[0]
                    for q,qq in wl.items():
                        wl[q] = ((1000,4000,0,1), qq[1])

        else:
            ax.clear()
            for i, s in enumerate(spectra1):
                ax.plot(self.wavenumber, s, linewidth=1,
                        label=label1 if not i else None, **args1)
            for i, s in enumerate(spectra2):
                ax.plot(self.wavenumber, s, linewidth=1,
                        label=label2 if not i else None, **args2)
            ax.legend()


    def draw_or_clear(self, ax):
        if self.spectra is not None and len(self.spectra) > 0:
            self.draw_ax(ax)
            if self.ltr == ax.xaxis_inverted():
                ax.invert_xaxis()
        else:
            ax.clear()

    def draw(self):
        self.draw_or_clear(self.ax)
        FigureCanvas.draw(self)
        if plt.fignum_exists(self.objectName()):
            self.popOut()

    def popOut(self):
        fig = plt.figure(self.objectName(), tight_layout=dict(pad=.6))
        fig.canvas.set_window_title(self.window_title)
        ax = fig.gca()
        self.draw_or_clear(ax)
        fig.canvas.draw_idle()
        if ax != self.popax:
            fig.show()
            self.popax = ax


class SpectraPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.window_title = 'Raw spectra'

    def draw_ax(self, ax):
        self.draw_ax_single(ax)

class NormPlotWidget(SpectraPlotWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.window_title = 'Normalized spectra'

class ACPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.window_title = 'Atmospheric correction'

    def draw_ax(self, ax):
        self.draw_ax_double(ax, 'Raw', 'Atm. corr.',
                            args2=dict(color='r', linestyle='--'))


class SCPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.window_title = 'Scattering correction'
        self.progressMode = False
        self.runSelected = None
        self.showSelected = None

    def setData(self, wavenumber, inspectra, spectra, runSelected, showSelected):
        self.wavenumber = wavenumber
        self.inspectra = inspectra
        self.spectra = spectra
        self.runSelected = runSelected
        self.setSelected(showSelected)

    def setSelected(self, sel):
        if self.runSelected is not None:
            if sorted(sel) == sorted(self.runSelected):
                self.showSelected = range(len(self.spectra))
            else:
                self.showSelected = []
        else:
            self.showSelected = sel
        self.draw_idle()
        self.updated.emit()

    def getSpectra(self):
        if self.spectra is None or self.runSelected is not None:
            return self.spectra
        return self.spectra[self.showSelected]

    # Rather ugly hack to show progress in the main window only: the draw method is
    # made to bypass the DataPlotWidget draw.
    def prepareProgress(self, wn):
        self.wavenumber = wn
        self.ax.clear()
        self.progressMode = True
        self.draw_idle()

    def endProgress(self):
        if self.progressMode:
            self.progressMode = False
            self.draw_idle()

    @pyqtSlot(np.ndarray, tuple)
    def progressPlot(self, curves, progress):
        if self.progressMode:
            self.ax.plot(self.wavenumber, curves.T,
                         color=plt.cm.jet(progress[0] / progress[1]), linewidth=1)
            self.draw_idle()

    def draw(self):
        if self.progressMode:
            return FigureCanvas.draw(self)
        return DataPlotWidget.draw(self)

    def draw_ax(self, ax):
        ax.clear()
        i = 0
        for s in self.showSelected:
            ax.plot(self.wavenumber, self.inspectra[s],
                    color='black', linewidth=1, label='Input' if not i else None)
            ax.plot(self.wavenumber, self.spectra[s],
                    linewidth=1, label='Corrected' if not i else None)
            i += 1
        if i:
            ax.legend()


class SGFPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        DataPlotWidget.__init__(self, parent=None)
        self.window_title = 'Savitzky-Golay filter'

    def draw_ax(self, ax):
        lines = ax.get_lines()
        nspec = len(self.spectra)
        if len(lines) == nspec * 2:
            for i, l in enumerate(lines):
                l.set_data(
                    self.wavenumber, self.inspectra[i] if i < nspec
                    else self.spectra[i - nspec])
            ax.relim()
            ax.autoscale_view()
        else:
            ax.clear()
            for i, s in enumerate(self.inspectra):
                ax.plot(self.wavenumber, s, linewidth=1,
                        color='b', label='Input' if not i else None)
            for i, s in enumerate(self.spectra):
                ax.plot(self.wavenumber, s, linewidth=1,
                        label='Smoothed' if not i else None)
            ax.legend()


class SRPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        DataPlotWidget.__init__(self, parent=None)
        self.window_title = 'Spectral region selection'
        self.mincut=0
        self.maxcut=10000
        self.cutwn=np.zeros(0)
        self.vspans = [None, None]

    def setData(self, wavenumber, inspectra, spectra, cutwn, mincut, maxcut):
        self.wavenumber = wavenumber
        self.inspectra = inspectra
        self.spectra = spectra
        self.mincut=mincut
        self.maxcut=maxcut
        self.cutwn=cutwn
        self.draw_idle()
        self.updated.emit()

    def getWavenumbers(self):
        return self.cutwn if self.cutwn is not None else self.wavenumber

    def draw_ax(self, ax):
        self.draw_ax_single(ax, spectra=self.inspectra)

        for v in self.vspans:
            if v is not None:
                v.remove()
        self.vspans = [None, None]
        mmwn = self.wavenumber[[0, -1]]
        mmwn = (min(mmwn), max(mmwn))
        if mmwn[0] < self.mincut:
            self.vspans[0] = ax.axvspan(mmwn[0], self.mincut, color='gray')
        if self.maxcut < mmwn[1]:
            self.vspans[1] = ax.axvspan(self.maxcut, mmwn[1], color='gray')


class BCPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        DataPlotWidget.__init__(self, parent=None)
        self.window_title = 'Baseline correction'

    def getSpectra(self):
        if self.spectra is not None:
            return self.inspectra - self.spectra
        return self.inspectra

    def draw_ax(self, ax):
        self.draw_ax_double(ax, 'Input', 'Baseline', args2=dict(color='k'))


class MCPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        DataPlotWidget.__init__(self, parent=None)
        self.window_title = 'mIRage correction'

    def draw_ax(self, ax):
        ss1 = self.inspectra / self.inspectra.max(1)[:, None]
        ss2 = self.spectra / self.spectra.max(1)[:, None]
        self.draw_ax_double(
            ax, spectra1=ss1, label1='Input', args1=dict(color='b'),
            spectra2=ss2, label2='Corrected', args2=dict(color='r'))

class MCInfoPlotWidget(FigureCanvas):
    def __init__(self, parent=None):
        self.fig = Figure(tight_layout=dict(pad=0.4))
        self.ax = self.fig.add_subplot(111)
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding,
                                   QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        self.logshifts = None # (nbreaks, npixels)
        self.breaks = None # (nbreaks, 2)
        # self.mode = 0

    # def mousePressEvent(self, event):
    #     self.mode = (self.mode + 1) % 2
    #     self.draw_idle()

    def setData(self, breaks, mcinfo):
        self.breaks = breaks
        if mcinfo is None:
            self.logshifts = None
            self.weights = None
        else:
            self.logshifts, self.weights = mcinfo
        self.draw_idle()

    def draw(self):
        ax = self.ax
        if self.logshifts is None:
            ax.clear()
        else:
            # if len(ax.collections) == len(self.logshifts):
            #     for i, l in enumerate(self.logshifts):
            #         ax.collections[i].set_offsets(np.column_stack(
            #             (self.weights[i]/self.weights[i].mean(), l)))
            #     ax.relim()
            #     ax.autoscale_view()
            # else:
            ax.clear()
            for i, l in enumerate(self.logshifts):
                ax.scatter(self.weights[i]/self.weights[i].mean(), l,
                           marker='x',
                           label='%g-%g' % tuple(self.breaks[i]))
                ax.set_ylabel(r'log$_{e}$ shift')
                ax.set_xlabel('Soft limit point reliability')
                ax.legend()
        # else:
            # lines = ax.get_lines()
            # logshifts = np.sort(self.logshifts, 1)
            # if len(lines) == len(logshifts):
            #     for i, l in enumerate(logshifts):
            #         lines[i].set_data(range(len(l)), l)
            #     ax.relim()
            #     ax.autoscale_view()
            # else:
            #     ax.clear()
            #     ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            #     ax.set_title(r'log$_{e}$ shift')
            #     for i, l in enumerate(logshifts):
            #         ax.plot(l, linewidth=1,
            #                      label='(%g - %g)' % tuple(self.breaks[i]))
            #     ax.legend()
        FigureCanvas.draw(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 24 23:44:05 2020

@author: carl
"""

import numpy as np

def normalize_spectra(method, y, wn=None, **kwargs):
    """
    Normalize the spectra in the matrix y (pixel,wavenum) with the given method

    Parameters
    ----------
    method : str
        Normalization method; one of: 'mean', 'area', 'wn', 'max', 'n2'
    y : array
        spectra in order (spectrum, wavenum)
    wn : array, optional
        wavenumber array; needed for the 'area' and 'wn' methods.
        The default is None.
    wavenum : float
        wavenumber to normalize at; used by 'wn'

    Raises
    ------
    ValueError
        if give an undefined method.

    Returns
    -------
    array
        Array of normalized spectra in the same format as y.
    """

    if method == 'mean':
        return (y.T / y.mean(axis=1)).T
    elif method == 'area':
         return (y.T / np.trapz(y, wn, axis=1)).T
    elif method == 'wn':
        idx = (np.abs(wn - kwargs['wavenum'])).argmin()
        return (y.T / y[:, idx]).T
    elif method == 'max':
        return (y.T / y.max(axis=1)).T
    elif method == 'n2':
        return (y.T / np.sqrt((y * y).mean(axis=1))).T
    raise ValueError('undefined normalization method ' + method)

def normalize_features(y, method):
    """
    Normalize intensities or other features.

    Parameters
    ----------
    y : array(nsamples, nfeatures)
        Data matrix, e.g. raw spectra or decomposition components.
    method : str
        One of 'none', 'mean1', 'mean0', 'mean0var1'

    Returns
    -------
    Normalized data.
    """
    if method == 'none':
        return y
    elif method == 'mean1':
        means = y.mean(0)
        means[means == 0] = 1
        return y / means
    elif method == 'mean0':
        return y - y.mean(0)
    elif method == 'mean0var1':
        y = y - y.mean(0)
        sds = y.std(0)
        sds[sds == 0] = 1
        return y / sds
    raise ValueError('undefined normalization method ' + method)




import matplotlib
matplotlib.use('QT5Agg')
import sys
from PyQt5 import uic
from PyQt5.QtWidgets import QApplication, QFileDialog, QErrorMessage
#from PyQt5.QtWidgets import QStyle, QProxyStyle, QMessageBox
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.Qt import QMainWindow, qApp
#from os.path import basename, dirname
import numpy as np
import scipy.signal, scipy.io
import matplotlib.pyplot as plt
import os.path
from imagedata import ImageData

Ui_MainWindow = uic.loadUiType("pixelpicker_ui.ui")[0]

class MyMainWindow(QMainWindow, Ui_MainWindow):

    closing = pyqtSignal()

    fileOptions = QFileDialog.Options() | QFileDialog.DontUseNativeDialog

    def __init__(self,parent=None):
        super(MyMainWindow, self).__init__(parent)
        qApp.installEventFilter(self)
        self.setupUi(self)

        self.data = ImageData()

        self.errorMsg = QErrorMessage(self)
        self.errorMsg.setWindowModality(Qt.WindowModal)

        self.pushButtonLoadWL.clicked.connect(self.loadWhite)
        self.pushButtonLoadSpectra.clicked.connect(self.loadSpectra)
        self.pushButtonLoad.clicked.connect(self.loadAnnotations)
        self.pushButtonSave.clicked.connect(self.saveAnnotations)

        self.pushButtonClear.clicked.connect(self.plot_visual.clearSelected)

#        self.comboBoxMethod.currentIndexChanged.connect()
        self.horizontalSliderWn.valueChanged.connect(self.wavenumberSlide)
        self.lineEditWn.editingFinished.connect(self.wavenumberEdit)

        self.comboBoxVisual.currentIndexChanged.connect(self.imageProjection)
        self.comboBoxCmaps.currentTextChanged.connect(self.plot_visual.setCmap)

        self.lineEditWn.setFormat("%.2f")
#        self.lineEditWavenumber.setRange(min=wmin, max=wmax, default=.5*(wmin+wmax))

        self.comboBoxAnnotation.currentIndexChanged.connect(self.plot_visual.setAnnotation)

        self.plot_visual.addedPoint.connect(self.plot_whitelight.addPoint)
        self.plot_visual.removedPoint.connect(self.plot_whitelight.removePoint)
        self.plot_visual.alteredCounts.connect(self.updateCounts)
        self.plot_whitelight.clickedPoint.connect(self.plot_visual.clickOnePoint)

    def closeEvent(self, event):
        self.closing.emit()
        plt.close('all')
        self.deleteLater()
        qApp.quit()

    def loadSpectra(self):
        "Load a file or return (error message, traceback) from trying"
        fileName, _ = QFileDialog.getOpenFileName(self,
                                                  "Load FTIR image", "",
                                                  "Matlab files (*.mat);;All files (*)",
                                                  options=MyMainWindow.fileOptions)
        self.data.readmat(fileName)
        self.plot_visual.setData(self.data.wavenumber, self.data.raw, self.data.wh)
        self.horizontalSliderWn.setMaximum(len(self.data.wavenumber) - 1)
        self.lineEditWn.setRange(self.data.wmin, self.data.wmax)
        self.imageProjection()
        self.plot_whitelight.load(os.path.splitext(fileName)[0]+'.jpg')

    def loadAnnotations(self):
        fileName, _ = QFileDialog.getOpenFileName(self,
                                                  "Load annotation map", "",
                                                  "Matlab files (*.mat);;All files (*)",
                                                  options=MyMainWindow.fileOptions)
        if fileName:
            s = scipy.io.loadmat(fileName)['annotations']
            self.plot_visual.setAnnotations(s)

    def saveAnnotations(self):
        fileName, _ = QFileDialog.getSaveFileName(self,
                                                  "Save annotation map", "",
                                                  "Matlab files (*.mat);;All files (*)",
                                                  options=MyMainWindow.fileOptions)
        if fileName:
            scipy.io.savemat(fileName, {'annotations': self.plot_visual.getAnnotations()})


    # Image visualization
    def loadWhite(self):
        fileName, _ = QFileDialog.getOpenFileName(self,
                                                  "Load white light image", "",
                                                  "Image files (*.jpg *.png);;All files (*)",
                                                  options=MyMainWindow.fileOptions)
        if fileName:
            self.plot_whitelight.load(fileName)

    def imageProjection(self):
        meth = self.comboBoxVisual.currentIndex()
        iswn = meth == 2
        self.lineEditWn.setEnabled(iswn)
        self.horizontalSliderWn.setEnabled(iswn)
        self.plot_visual.setProjection(meth, -1-self.horizontalSliderWn.value())

    def updateCounts(self, counts):
        for i in range(self.comboBoxAnnotation.count()):
            t = self.comboBoxAnnotation.itemText(i)
            p = t.rfind('(')
            t = t[:p] if p >= 0 else t + ' '
            self.comboBoxAnnotation.setItemText(i, t + '(%d)' % counts[i])

    def sliderToBox(self, slider, box, ixfinder):
        wn = self.plot_visual.getWavenumbers()
        if wn is not None:
            if (not box.hasAcceptableInput() or
                    slider.value() != ixfinder(wn, box.value())):
                box.setValue(wn[-1-slider.value()])
        elif (not box.hasAcceptableInput() or
                slider.value() != int(round(slider.maximum() *
                (box.value() - self.wmin) / (self.wmax - self.wmin)))):
            box.setValue(self.wmin + (self.wmax - self.wmin) *
                         slider.value() / slider.maximum())

    def boxToSlider(self, slider, box, ixfinder):
        wn = self.plot_visual.getWavenumbers()
        if wn is not None:
            if box.hasAcceptableInput():
                slider.setValue(ixfinder(wn, box.value()))
            else:
                box.setValue(wn[-1-slider.value()])
        else:
            slider.setValue(int(round(slider.maximum() *
                  (box.value() - self.wmin) / (self.wmax - self.wmin))))

    def wavenumberSlide(self):
        self.sliderToBox(self.horizontalSliderWn, self.lineEditWn,
                         lambda wn, val: len(wn)-1 - (np.abs(wn - val)).argmin())
        self.plot_visual.setProjection(2, -1-self.horizontalSliderWn.value())

    def wavenumberEdit(self):
        self.boxToSlider(self.horizontalSliderWn, self.lineEditWn,
                         lambda wn, val: len(wn)-1 - (np.abs(wn - val)).argmin())




if __name__ == '__main__':
    app = QApplication.instance()
    if not app:
        app = QApplication(sys.argv)
    window = MyMainWindow()
    window.show()
    sys.exit(app.exec_())




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 14:01:08 2020

@author: carl
"""

import os.path
import numpy as np
import scipy.signal, scipy.io
import pandas as pd
from pymatreader import read_mat
import h5py

from .opusreader import OpusReader
from .ptirreader import PtirReader
from .omnicreader import OmnicReader
from ..algorithms.util import pixels_fit
from .. import octavvs_version

class SpectralData:
    """
    The list of current files and the raw data loaded from a hyperspectral
    image file in the GUIs.
    """
    def __init__(self):
        self.foldername = '' # Root dir
        self.filenames = []  # With full paths
        self.curFile = ''    # The currently loaded file
        self.wavenumber = None # array in order low->high
        self.wmin = 800
        self.wmax = 4000
        self.raw = np.empty((0,0)) # data in order (pixel, wavenumber)
        self.wh = (0, 0)  # Width and height in pixels
        self.pixelxy = None # Positions of pixels in image
        self.images = None # list of Image loaded from raw data file
        self.filetype = None # str from [txt, mat, opus, ptir]

    def set_width(self, w):
        try:
            w = int(w)
        except ValueError:
            w = 0
        if w <= 0:
            return False
        h = (self.raw.shape[0] + w - 1) // w
        # if w * h != self.raw.shape[0]:
        #     return False
        self.wh = (w, h)
        return True

    def set_height(self, h):
        if self.set_width(h):
            self.wh = (self.wh[1], self.wh[0])
            return True
        return False

    def read_matrix(self, filename):
        """
        Read data from a file, with some error checking. The object is modified
        only if the file is successfully loaded.
        """
        wh = None
        xy = None
        images = None
        raw = None
        fext = os.path.splitext(filename)[1].lower()
        if fext in ['.txt', '.csv', '.mat', '.xls', '.xlsx']:
            s = {} # possible
            if fext == '.mat':
                filetype = 'mat'
                try:
                    s = read_mat(filename)
                except TypeError:
                    # Workaround for uint16_codec bug (pymatreader
                    # assumes mio5, not mio4)
                    s = scipy.io.loadmat(filename)
                # Assume data are in the biggest matrix in the file
                ss = max(s.items(), key=lambda k: np.size(k[1]) )[1]
            elif fext in ['.xls', '.xlsx']:
                filetype = 'xls'
                ss = pd.read_excel(filename, header=None).to_numpy()
            else:
                filetype = 'txt'
                ss = pd.read_csv(filename, sep=None, engine='python',
                                 header=None).to_numpy()
            # Check for some particular format(s)
            if 'ImR' in s and 'lambda' in s:
                # Matlab file with data in ImR, wavelengths in lambda
                # (encountered with hyperspectral data in Lund)
                ss = s['ImR']
                if ss.ndim == 2:
                    raw = ss
                else:
                    assert ss.ndim == 3
                    wh = [ss.shape[1], ss.shape[0]]
                    raw = ss.reshape((-1, ss.shape[2]))
                wn = s['lambda'].flatten()
                assert len(wn) == raw.shape[1]
            elif ss.ndim == 2 and ss.shape[0] > 1 and ss.shape[1] > 1:
                if 'wh' in s:
                    wh = s['wh'].flatten()
                # Assume wavenumbers are a) first row or b) first column
                for transp in range(2):
                    if transp:
                        ss = ss.T
                    wn = ss[0]
                    cols = None
                    # if mixed data, remove strings (these should be saved
                    # as annotations!)
                    if wn.dtype == object:
                        cols = [not isinstance(s, str) for s in wn]
                        wn = wn[cols].astype(float, casting='unsafe')
                        if len(wn) < 2:
                            continue
                    deltas = np.diff(wn)
                    if np.all(deltas > 0) or np.all(deltas < 0):
                        if cols is not None:
                            raw = ss[1:, cols].astype(float)
                        else:
                            raw = ss[1:]
                        break
                if raw is None:
                    raise RuntimeError(
                        'first column or row must contain sorted wavenumbers')
            else:
                raise RuntimeError(
                    'file does not appear to describe an FTIR image matrix')
        else:
            if fext == '.ptir' or fext == '.hdf':
                filetype = 'ptir'
                reader = PtirReader(filename)
                xy = reader.xy
            elif fext in ['.spa', '.spg']:
                filetype = 'omnic'
                reader = OmnicReader(filename)
            else:
                filetype = 'opus'
                reader = OpusReader(filename)
            raw = reader.AB
            wn = reader.wavenum
            wh = reader.wh
            images = reader.images

        # Ensure ascending wavenumbers
        if wn[0] > wn[1]:
            wn = wn[::-1]
            raw = raw[:, ::-1]

        diffsign = np.sign(np.diff(wn))
        if (diffsign >= 0).any() and (diffsign <= 0).any():
            raise RuntimeError('wavenumbers must be sorted')
        npix = raw.shape[0]
        if wh is not None:
            if len(wh) != 2:
                raise RuntimeError('Image size in "wh" must have length 2')
            wh = (int(wh[0]), int(wh[1]))
            if not pixels_fit(npix, wh):
                raise RuntimeError('Image size in "wh" does not match data size')
            self.wh = wh
        elif npix != self.wh[0] * self.wh[1]:
            res = int(np.sqrt(npix))
            if npix == res * res:
                self.wh = (res, res)
            else:
                self.wh = (npix, 1)
        self.raw = raw
        self.wavenumber = wn
        self.wmin = self.wavenumber.min()
        self.wmax = self.wavenumber.max()
        self.images = images
        self.pixelxy = xy
        self.curFile = filename
        self.filetype = filetype

    def save_matrix_matlab_ab(self, filename, wn, ydata):
        ab = np.hstack((wn[:, None], ydata.T))
        abdata = {'AB': ab, 'wh': self.wh }
        if self.pixelxy is not None:
            abdata['xy'] = self.pixelxy
        scipy.io.savemat(filename, abdata)

    def save_matrix_matlab_quasar(self, filename, wn, ydata):
        out = {'y': ydata, 'wavenumber': wn}
        if self.pixelxy is not None:
            map_x = np.array([x for (x,y) in self.pixelxy])
            map_y = np.array([y for (x,y) in self.pixelxy])
        else:
            map_x = np.tile(self.wh[0], self.wh[1])
            map_y = np.repeat(range(self.wh[0]), self.wh[1])
        out['map_x'] = map_x[:, None]
        out['map_y'] = map_y[:, None]
        scipy.io.savemat(filename, out)

    def save_matrix_ptir(self, filename, wn, ydata):
        f = h5py.File(filename, mode='w')

        f.attrs['DocType'] = b'IR'
        f.attrs['SoftwareVersion'] = octavvs_version.encode('utf8')

        digits = int(np.log10(len(ydata))) + 1
        for i, d in enumerate(ydata):
            mm = f.create_group(f'Measurement{i+1:0{digits}}')
            if self.pixelxy is not None:
                x, y = self.pixelxy[i]
            else:
                x = i % self.wh[0]
                y = i // self.wh[0]
            mm.attrs['LocationX'] = [x]
            mm.attrs['LocationY'] = [y]

            if not i:
                specval = mm.create_dataset('Spectroscopic_Values',
                                            data=wn[None, :])

            ch = mm.create_group('Channel_000')
            rd = ch.create_dataset('Raw_Data', data=d[None, :])
            rd.attrs['Spectroscopic_Values'] = specval.ref

        if self.images is not None and len(self.images):
            imgs = f.create_group('Images')
            digits = max(3, int(np.log10(len(self.images))) + 1)
            for i, im in enumerate(self.images):
                imds = imgs.create_dataset(f'Image_{i:0{digits}}',
                                           data=im.data[::-1])
                imds.attrs['Label'] = np.bytes_(im.name)
                imds.attrs['PositionX'] = [im.xy[0]]
                imds.attrs['PositionY'] = [im.xy[1]]
                imds.attrs['SizeWidth'] = [im.wh[0]]
                imds.attrs['SizeHeight'] = [im.wh[1]]

        f.close()

    def save_matrix(self, filename, fmt='ab', wn=None, ydata=None):
        """
        Save processed data in some format.

        Parameters
        ----------
        filename : str
            Name of the output file.
        fmt : str, optional
            File format: 'ab', 'quasar' or 'ptir'. The default is 'ab', a
            MATLAB file with array 'AB' with wavenumbers in the first _column_,
            image size in 'wh' and optionally pixel coordinates in 'xy'.
            The 'quasar' format has wavenumbers in a separate array.
            The 'ptir' format is a reduced variant of the HDF5 format used
            by PTIR Studio.
        wn : 1D array, optional
            Wavenumber vector. The default is self.wn.
        ydata : 2D array, optional
            Data in (pixel, wn) order. The default is self.raw.

        Returns
        -------
        None.

        """
        if wn is None:
            wn = self.wn
        if ydata is None:
            ydata = self.raw
        if fmt == 'ab':
            self.save_matrix_matlab_ab(filename, wn, ydata)
        elif fmt == 'quasar':
            self.save_matrix_matlab_quasar(filename, wn, ydata)
        elif fmt == 'ptir':
            self.save_matrix_ptir(filename, wn, ydata)
        else:
            raise ValueError('Unknown save file format')





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Carl Troein

Atmospheric and scattering correction

"""

import gc
from time import monotonic
import numpy as np
import sklearn.linear_model
import sklearn.cluster
#import statsmodels.multivariate.pca
from scipy.interpolate import PchipInterpolator
from scipy.signal import hilbert
#from scipy.io import loadmat, savemat

from pymatreader import read_mat
from pkg_resources import resource_filename
from scipy.interpolate import RectBivariateSpline

import matplotlib
matplotlib.use('QT5Agg')
import matplotlib.pyplot as plt

from .util import pca_nipals, find_wn_ranges

def kkre(wn, ref):
    wn2 = wn ** 2.
    wa = wn * ref
    kk = np.empty_like(wn)
    for i in range(len(wn)):
        with np.errstate(divide='ignore', invalid='ignore'):
            fg = wa / (wn2 - wn[i] ** 2.)
        if i == 0 or i == len(wn) - 1:
            fg[i] = 0
        else:
            fg[i] = (fg[i-1] + fg[i+1]) / 2
        kk[i] = 2/np.pi * np.trapz(x=wn, y=fg)
    if wn[0] < wn[-1]:
        return kk
    return -kk

def hilbert_n(wn, ref, zeropad=500):
    """
    Compute the Kramers-Kronig relations by Hilbert transform, extending the absorption spectrum
    with 0 to either size and resampling the spectrum at evenly spaced intervals if it is found to
    be unevenly sampled.
    """
    # Cache some data structures to avoid having to reinitialize them on every call.
    if not hasattr(hilbert_n, "wn") or hilbert_n.wn is not wn or hilbert_n.zp != zeropad:
        hilbert_n.wn = wn
        even = (wn[-1] - wn[0]) / (len(wn) - 1)
        diff = np.abs((np.diff(wn) - even) / even).mean()
        hilbert_n.zp = zeropad
        hilbert_n.evenspaced = diff < 1e-3
        hilbert_n.increasing = wn[0] < wn[-1]
#        print('hilbert',hilbert_n.evenspaced,hilbert_n.increasing)
        if not hilbert_n.evenspaced:
            hilbert_n.lin = np.linspace(min(wn[0], wn[-1]), max(wn[0], wn[-1]), len(wn))
        hilbert_n.npad = int(len(wn) / abs(wn[-1] - wn[0]) * zeropad)
        hilbert_n.nim = np.zeros((len(wn) + 2 * hilbert_n.npad))

    if hilbert_n.evenspaced:
        if hilbert_n.npad == 0:
            if hilbert_n.increasing:
                hilbert_n.nim = ref
            else:
                hilbert_n.nim = ref[::-1]
        elif hilbert_n.increasing:
            hilbert_n.nim[hilbert_n.npad:hilbert_n.npad+len(wn)] = ref
        else:
            hilbert_n.nim[hilbert_n.npad:hilbert_n.npad+len(wn)] = ref[::-1]
    else:
        if hilbert_n.increasing:
            hilbert_n.nim[hilbert_n.npad:hilbert_n.npad+len(wn)] = \
                PchipInterpolator(wn, ref)(hilbert_n.lin)
        else:
            hilbert_n.nim[hilbert_n.npad:hilbert_n.npad+len(wn)] = \
                PchipInterpolator(wn[::-1], ref[::-1])(hilbert_n.lin)
    nreal = -np.imag(hilbert(hilbert_n.nim))
    if hilbert_n.npad:
        nreal = nreal[hilbert_n.npad:-hilbert_n.npad]
    if hilbert_n.evenspaced:
        return nreal if hilbert_n.increasing else nreal[::-1]
    return PchipInterpolator(hilbert_n.lin, nreal)(wn)

def compute_model(wn, ref, model='konevskikh', n_components=7,
                  a=np.linspace(1.1, 1.5, 10), d=np.linspace(2, 7.1, 10),
                  bvals=10, constantcomponent=True,
                  linearcomponent=False, variancelimit=None):
    """
    Compute the extinction matrix for resonant Mie scattering, then PCA transform
    it. The model components will all be orthogonal to the reference
    Parameters:
    wn: array of wavenumbers
    ref: reference spectrum
    n_components: number of PCA components to use
    a: array of values for the parameter a (index of refraction)
    d: array of values for the parameter d (sphere size, units µm)
    bvals: number of values for parameter b (mixing of a and real part of n from absorption ref)
    model: 'bassan': original 3-parameter model by Bassan et al.
        'konevskikh': faster and more accurate 2-param model by Konevskikh et al.
        'rasskazov': cylindrical model by Rasskazov et al.
    constantcomponent: include a constant component in the EMSC model
    linearcomponent: include a component proportional to wn
    variancelimit: if not None, use as many PCA components as needed to
       explain this fraction of the variance of the extinction matrix (around 0.9996)
    """

    if model == 'konevskikh':
        nim = ref / ref.max() / (wn * 100)
        # nim = ref / (wn * 100)
        nre = hilbert_n(wn, nim, 300)
        nmin = nre.min()
        # print('refmax', ref.max(), 'nre', nmin, nre.max(), 'nim', nim.min(), nim.max())
        if nmin < -1:
            nre = nre / -nmin
            nim = nim / -nmin
        # d in m times the factor 4*pi
        dd = d * (4e-6 * np.pi)
        # My revised distribution of alpha_0 and gamma
        alpha_0 = np.linspace(dd[0] * (a[0]-1), dd[-1] * (a[-1]-1), len(a))
        gamma = .25 * 2 * np.log(10) / np.pi * np.linspace(1 / alpha_0[0], 1 / alpha_0[-1], len(alpha_0))
        # Solheim's distributions of alpha_0 and gamma
        # alpha_0 = dd * (a - 1)
        # gamma = .25  * 2 * np.log(10) / np.pi / alpha_0

        Q = np.empty((len(alpha_0) * len(gamma), len(wn)))  # Initialize the extinction matrix

#        print('alpha_0', alpha_0)
#        print('gamma', gamma)
        # Build the extinction matrix
        n_row = 0
        # maxrho = 0
        # maxbeta = 0
        for a0 in alpha_0:
            for g in gamma:
                rho = a0 * (1. + g * nre) * wn * 100
                denom = 1. / g + nre
                tanbeta = nim / denom
                beta = np.arctan2(nim, denom)
                cosb = np.cos(beta)
                cosbrho = cosb / rho
                # maxbeta = max(maxbeta, beta.max())
                # maxrho = max(maxrho, rho.max())

                # Following Konevskikh et al 2016
                Q[n_row] = 2. - 4. * cosbrho * (np.exp(-rho * tanbeta) *
                     (np.sin(rho - beta) + cosbrho * np.cos(rho - 2 * beta)) -
                     cosbrho * np.cos(2 * beta))

                n_row += 1
        # print('maxbeta', maxbeta, 'maxrho', maxrho)
#        savemat('everything-p.mat', locals())
    elif model == 'bassan':
        # I'm a bit confused about the division/multiplication by wn.
        # Bassan's matlab code uses wn*ref in kkre
            # bassan2010 says k (=n_im) is proportional to ref
            # but various sources say k is ref/wn * const
        # My kkre reproduces's Bassan with the same wn*ref
        # My hilbert_n gives the same output with Hilbert transform of ref
        # Causin's python code Hilbert transforms ref for Bassan's algorith but ref/wn for Konevskikh's!
        # Solheim's matlab code Hilbert transforms ref/wn

        nkk = kkre(wn, ref / wn)
#        nkk = hilbert_n(wn, ref / wn, 300)
        nkk = nkk / abs(nkk.min())

        # Build the extinction matrix
        Q = np.empty((len(a) * bvals * len(d), len(wn)))  # Initialize the extinction matrix
        n_row = 0
        # d in cm times the factor 4*pi
        dd = d * (4e-4 * np.pi)
        for i in range(len(a)):
            b = np.linspace(0.0, a[i] - 1.01, bvals)  # Range of amplification factors of nkk
            for j in range(len(b)):
                n = a[i] + b[j] * nkk  # Compute the real refractive index
                for k in range(len(dd)):
                    rho = dd[k] * (n - 1.) * wn
                    #  Compute the extinction coefficients for each combination of a, b and d:
                    Q[n_row] = 2. - 4. / rho * np.sin(rho) + \
                               4. / (rho * rho) * (1. - np.cos(rho))
                    n_row += 1
    elif model == 'rasskazov':
        nim = ref / ref.max() / (wn * 100)
        nre = hilbert_n(wn, nim, 300)
        nmin = nre.min()
        if nmin < -1:
            nre = nre / -nmin
            nim = nim / -nmin

        if not hasattr(compute_model, 'qtable'):
            qtable = read_mat(resource_filename(
                'octavvs.reference_spectra', "Q_table.mat"))
            compute_model.qtable = RectBivariateSpline(
                qtable['x_table'], qtable['n_table'],
                qtable['Qsca_table'])

        dd = d * (1e-4 * np.pi)
        Q = np.empty((len(a) * bvals * len(dd), len(wn)))
        n_row = 0
        for i in range(len(a)):
            b = np.linspace(0.0, a[i] - 1.01, bvals)
            for bb in b:
                n = a[i] + bb * nre
                for r in dd:
                    x = r * wn
                    Q[n_row] = compute_model.qtable(x, n, grid=False)
    else:
        raise ValueError("model must be one of 'bassan', 'konevskihk' or 'rasskazov'")

    n_nonpca = 1 + bool(constantcomponent) + bool(linearcomponent)

    # Orthogonalization of the model to improve numeric stability
    refn = ref / np.sqrt(ref@ref)
    Q = Q - (Q @ refn)[:,None] @ refn[None,:]

    # Perform PCA of the extinction matrix
    pca = pca_nipals(Q, ncomp=n_components, tol=1e-5, copy=False, explainedvariance=variancelimit)
    model = np.empty((n_nonpca + pca.shape[0], pca.shape[1]))
    model[n_nonpca:, :] = pca

#    # This method is up to 50% slower but gives additional info
#    pca = statsmodels.multivariate.pca.PCA(Q, ncomp=n_components,
#           method='nipals', tol=1e-5, demean=False, standardize=False)
#    esum = pca.eigenvals.cumsum()
#    n_components = np.min((np.searchsorted(esum, esum[-1] * variancelimit) + 1, len(esum)))
#    model = np.zeros((n_nonpca + n_components, len(wn)))
#    model[n_nonpca:, :] = pca.loadings[:,:n_components].T

    model[0,:] = ref
    # These model components will not be orthogonal to the PCA components
    if constantcomponent:
        model[1,:] = 1 - refn.sum() * refn
    if linearcomponent:
        lin = np.linspace(0., 1., len(wn))
        model[n_nonpca-1,:] = lin - (lin @ refn[:,None]) * refn
#        for i in range(1, n_nonpca):
#            model[i,:] = model[i,:] - refn * (model[i,:] @ refn)
#        for i in range(1, n_nonpca):
#            w = model[i, :] - np.sum(np.dot(model[i, :], b) * b for b in model[0:i, :])
#            model[i,:] = w / np.sqrt(w @ w)

#    savemat('everything_after-p.mat', locals())

    # Orthogonalization of the model to improve numeric stability (doing it after PCA is only
    # marginally slower)
#    for i in range(len(model)):
#        v = model[i, :]
#        w = v - np.sum(np.dot(v, b) * b for b in model[0:i, :])
#        model[i,:] = w / np.linalg.norm(w)

#    print('model',ref.sum(), model.sum())
    return model

def stable_rmiesc_clusters(iters, clusters):
    """
    Make a cluster size scheme for reliable convergence in rmiesc.
    Parameters:
    iters: The number of basic iterations to be used, preferably at least about 12-20
    Returns:
    array of cluster sizes (or zeros) for each iteration; this will be bigger than
    the input iterations
    """
    iters = max(2, iters) * 2
    cc = np.zeros(iters, dtype=np.int)
    cc[:iters//3] = 1
    cc[iters//2:iters*3//4] = clusters
#    cc = np.zeros(max(2, iters) * 3 // 2, dtype=np.int)
#    cc[:iters] = clusters
    return cc

def solve_model(qmodel, app, individual_references=None,
                weights=None, scaledzeroweights=None,
                prefit_reference=False):
    """
    Solve the EMSC least squares problem for one or more spectra.
    Parameters:
    qmodel: EMSC model; (component, wn)
    app: Apparent spectra; (N, wn)
    individual_references: corrected spectra from previous iteration, only
    needed if different from the current reference; (N, wn)
    weights: Array of weights for the fitting; (wn) or None
    scaledzeroweights: penalty on nonzero solutions, to be applied in regions
    where the reference is near-zero. Array of sqrt(1+lamba**2) where lambda
    may be a constant >0 in near-zero regions and 0 elsewhere.
    prefit_reference: if True, fit reference by projection before fitting
    the linear model; this vastly improves stability
    """
    if weights is not None:
        qmodel = qmodel * weights
        app = app * weights
    elif scaledzeroweights is not None:
        qmodel[1:, :] = qmodel[1:, :] * scaledzeroweights
        # if not prefit_reference:
        #     smod[0, :] = qmodel[0, :]
        app = app * scaledzeroweights

    if individual_references is not None:
        w = (app * individual_references).sum(1) / (
            individual_references * individual_references).sum(1)
        projs = w[:,None] * individual_references
        app = app - projs

        if prefit_reference:
            cons = np.linalg.lstsq(qmodel[1:, :].T, app.T, rcond=None)[0]
            corrected = app - cons.T @ qmodel[1:, :]
        else:
            cons = np.linalg.lstsq(qmodel.T, app.T, rcond=None)[0]
            corrected = app - cons[1:, :].T @ qmodel[1:, :]

        corrected = corrected + projs
        residuals = ((corrected - individual_references)**2).sum(1)
        # print('resids', residuals.mean(), w.mean(), cons[0].mean(), cons[1].mean())
    else:
        cons = np.linalg.lstsq(qmodel.T, app.T, rcond=None)[0]
        corrected = app - cons[1:, :].T @ qmodel[1:, :]
        residuals = ((corrected - qmodel[0, :])**2).sum(1)
        # print('resids', residuals, cons[0], cons[1])

    return cons, corrected, residuals


def compute_and_solve_model(wn, app, ref, modelparams, solvingparams,
                            individual_references=None, plot=None):
    """
    Returns:
        corrected: spectra
        residuals: .
    """
    onedim = np.ndim(app) == 1
    if onedim:
        app = app[None, :]

    qmodel = compute_model(wn, ref, **modelparams)
    cons, corrected, residuals = solve_model(
        qmodel, app, individual_references=individual_references,
        **solvingparams)

    # if plot:
    #     fig = plt.figure('model components', tight_layout=dict(pad=.6))
    #     ax = fig.gca()
    #     cs = cons[:, 0] if cons.ndim > 1 else cons
    #     ax.plot(wn, qmodel.T * cs,
    #             color=plt.cm.jet(plot[0]/plot[1]), linewidth=1)
    #     fig.canvas.draw_idle()

    #     w = (app * ref).sum(app.ndim - 1) / (ref * ref).sum()
    #     projs = (w[:,None] if w.ndim else w) * ref
    #     fig = plt.figure('spectrum sans ref', tight_layout=dict(pad=.6))
    #     ax = fig.gca()
    #     ax.plot(wn, (app-projs).T, color=plt.cm.jet(plot[0]/plot[1]), linewidth=1)
    #     ax.plot(wn, projs.T, linewidth=1, c='k')
    #     fig.canvas.draw_idle()

    #     fig = plt.figure('component weights', tight_layout=dict(pad=.6))
    #     ax = fig.gca()
    #     # print('cons', cs)
    #     ax.plot(cs, color=plt.cm.jet(plot[0]/plot[1]), linewidth=1)
    #     fig.canvas.draw_idle()

    if onedim:
        return corrected[0], residuals[0]
    return corrected, residuals

def rmiesc(wn, app, ref, iterations=10, clusters=None, modelparams={},
           weighted=False, zeroregionpenalty=None,
           prefit_reference=True, autoiterations=False,
           targetrelresiduals=0.95, verbose=False,
           progressCallback=None, progressPlotCallback=None):
    """

    Correct scattered spectra using Bassan's algorithm. This implementation does no orthogonalization
    of the extinction matrix or PCA components relative to the reference, nor is the reference smoothed
    or filtered through a sum of gaussians as in the original Matlab implementation.

    Parameters
    ----------
    wn : array
        sorted array of wavenumbers (high-to-low or low-to-high)
    app : array
        apparent spectrum, shape (pixels, wavenumbers)
    ref : array
        reference spectrum; array (wavenumbers)
    iterations : int
        number of iterations of the algorithm
    clusters : int or list
        If not None, cluster pixels into this many clusters in each iteration
        and use a common reference spectrum for each cluster. May be given as
        a list with one value per iteration, in which case 0 means to reuse
        clusters from the previous iteration and mix new/old references for
        stable convergence. If clusters is negative, use stable_rmiesc_clusters
        to generate the list.
    modelparams : dict
        Parameters used in compute_model.

        n_components :
            number of principal components to be calculated for
            the extinction matrix.
        a :
            indexes of refraction to use in model.
        d :
            sphere sizes to use in model, in micrometers.
        bvals :
            number of values for the model parameter b.
        model :
            model to use for the extinction matrix; see compute_model.
        linearcomponent :
            if True, include a linear term in the model(used in Bassan's paper only).
        variancelimit :
           if a number (around 0.9996), use as many PCA components
           as needed to explain this fraction of the variance of the extinction matrix.
    weighted : bool
        downweight the 1800 to 2800 cm-1 region when fitting
    zeroregionpenalty : bool
        penalize regions where the reference is near-zero
    prefit_reference : bool
        vastly improve stability by fitting reference before scattering
    autoiterations : bool
        if True, iterate until residuals stop improving
    targetrelresiduals : float
        if autoiterations, stop when this relative change in residuals is seen
    verbose : bool
        print progress information
    progressCallback : func(int a, int b)
        callback function called to indicate that the processing is complete
        to a fraction a/b.
    progressPlotCallback : func(array ref, (a, b))
        like the above with a (cluster) reference spectrum

    Returns
    -------
    corrected apparent spectra (the best encountered if autoiterations, else the final ones)

    """


    # The input can be a single spectrum or a matrix of spectra. If the former, squeeze at the end.
    squeeze = False
    if app.ndim == 1:
        app = app[None,:]
        squeeze = True

    if np.isscalar(clusters):
        if clusters == 0:
            clusters = None
        elif clusters < 0:
            clusters = stable_rmiesc_clusters(iterations, -clusters)
            iterations = len(clusters)
        else:
            clusters = np.repeat(clusters, iterations)
    elif clusters is not None:
        if len(clusters) != iterations:
            raise ValueError('len(clusters) must match iterations')

    if weighted:
        weights = np.ones_like(wn)
        weights[range(*find_wn_ranges(wn, [[1800, 2800]])[0])] = .001 ** .5
        # weights = weights[:, None]
    else:
        weights = None
    solvingparams = dict(weights=weights,
                         prefit_reference=prefit_reference)
    if zeroregionpenalty:
        solvingparams['scaledzeroweights'] = np.sqrt(
            1 + ((ref < ref.max() * .01) * zeroregionpenalty) ** 2)
    else:
        solvingparams['scaledzeroweights'] = None

    if progressCallback:
        # Compute the number of progress steps
        progressA = 0
        if clusters is None:
            progressB = 1 + (iterations > 1) * len(app)
        else:
            progressB = 1
            prev = 1
            for cl in clusters[1:]:
                if cl > 0:
                    prev = cl
                progressB += prev
    startt = monotonic()

    corrected = None # Just to get rid of warnings in the editor; will be set on iteration 0

    # Set parameters for automatic iteration control
    autoupadd = 1 # Residual going up counts as residual going down too little this many times
    automax = 5   # Stop when residual has gone down too little this many times


    if clusters is not None:
        # Cluster mode: In each iteration, after correcting all the spectra, cluster them.
        # Then take the mean of the corrected spectra in each cluster as the new reference
        # for that cluster in the next iteration. Spectra should still have their own reference
        # in the EMSC step. To avoid having to loop over all the spectra in a cluster when
        # solving that equation system, we remove the projection of the reference from the
        # apparent spectra.

        if progressPlotCallback:
            progressPlotCallback(ref, (0, iterations))

        corrected, residuals = compute_and_solve_model(
            wn, app, ref,
            modelparams=modelparams, solvingparams=solvingparams)
        ref = []    # Force init
        if autoiterations:
            unimproved = np.zeros(len(app), dtype=int)

        if progressCallback:
            progstep = 1 # Current progress bar step size
            progressA += progstep
            progressCallback(progressA, progressB)

        for iteration in range(1, iterations):
            corrected[corrected < 0] = 0

            curc = clusters[iteration] # Current cluster size
            if curc > 0:
                progstep = curc

            # Possibly recluster the spectra and compute reference spectra
            if curc > 0:
                # Find the remaining spectra to process
                if autoiterations:
                    if curc != clusters[iteration-1]:
                        unimproved = np.zeros(len(app), dtype=int)
                        nds = len(app)
                        notdone = range(nds)
                    else:
                        notdone = np.where(unimproved <= automax)[0]
                        nds = len(notdone)
                    # Skip this iteration if every spectrum has stopped improving
                    if nds == 0:
                        progressA += progstep
                        if progressCallback:
                            progressCallback(progressA, progressB)
                        continue
                else:
                    nds = len(app)
                    notdone = range(nds)

                # Find the clusters of spectra
                curc = min(curc, nds)
                if curc == nds:
                    # Each index in its own little array
                    cindices = [np.array([i]) for i in notdone]
                elif curc > 1:
                    kmeans = sklearn.cluster.MiniBatchKMeans(curc)
                    labels = np.zeros(len(app)) - 1
                    labels[notdone] = kmeans.fit_predict(corrected[notdone,:])
                    # Quadratic but simple
                    cindices = [np.where(labels == i)[0] for i in range(curc)]
                else:
                    # One big cluster
                    cindices = [np.array(notdone)]

                # Compute cluster references
                if(len(ref) != curc):
                    ref = np.zeros((curc, len(wn)))
                empties = 0
                for cl in range(curc):
                    if len(cindices[cl]):
                        ref[cl,:] = corrected[cindices[cl]].mean(0)
                    else:
                        empties += 1
                if empties:
                    print('Info: %d empty cluster(s) at iteration %d' % (empties, iteration))
            else:
                # Mix old reference and corrected spectrum. This requires the clusters
                # to remain unchanged.
                for cl in range(len(ref)):
                    if len(cindices[cl]):
                        # Stop processing clusters only when all members are done
                        if autoiterations and (unimproved[cindices[cl]] > automax).all():
                            cindices[cl] = []
                        else:
                            ref[cl,:] = .5 * corrected[cindices[cl]].mean(0) + .5 * ref[cl,:]

            if progressPlotCallback:
                progressPlotCallback(ref, (iteration, iterations))

            for cl in range(len(ref)):
                ix = cindices[cl] # Indexes of spectra in this cluster
                if len(ix) > 0 and autoiterations:
                    ix = ix[unimproved[ix] <= automax]
                if len(ix) == 0:
                    continue

                corrs, resids = compute_and_solve_model(
                    wn, app[ix], ref[cl], individual_references=corrected[ix],
                    modelparams=modelparams, solvingparams=solvingparams)

                improved = resids < residuals[ix]

                if autoiterations:
                    iximp = ix[improved]  # Indexes of improved spectra
                    impmore = resids[improved] < residuals[iximp] * targetrelresiduals
                    unimproved[iximp[impmore]] = 0
                    unimproved[iximp[np.logical_not(impmore)]] += 1
                    unimproved[ix[np.logical_not(improved)]] += autoupadd

                    corrected[iximp, :] = corrs[improved, :]
                    residuals[iximp] = resids[improved]
#                    corrected[ix,:] = corrs
#                    residuals[ix] = resids
                else:
                    corrected[ix,:] = corrs
                    residuals[ix] = resids

                nimprov = improved.sum()

                if verbose:
                    print("iter %3d, cluster %3d (%5d px): avgres %8.3g  imprvd %4d  time %f" %
                          (iteration, cl, len(ix), resids.mean(), nimprov, monotonic()-startt))
                if progressCallback:
                    progressCallback(progressA + cl + 1, progressB)
            if progressCallback:
                progressA += progstep
                progressCallback(progressA, progressB)
            gc.collect()    # Because my old laptop was unhappy with RAM usage otherwise

    else:
        ref = ref.copy()
        ref[ref < 0] = 0
        # For efficiency, compute the model from the input reference spectrum only once
        corrected, residuals = compute_and_solve_model(
            wn, app, ref,
            modelparams=modelparams, solvingparams=solvingparams,
            plot=(0, iterations))

        if progressPlotCallback:
            progressPlotCallback(ref, (0, len(app) + 1))

        if verbose:
            print("all pixels, iter %2d  time %f" % (0, monotonic()-startt))
        if progressCallback:
            progressA += 1
            progressCallback(progressA, progressB)
        if iterations > 1:
            for s in range(len(app)):
                gc.collect()
                unimproved = 0
                ref = corrected[s, :]  # Corrected spectrum as new reference
                for iteration in range(1, iterations):
                    ref[ref < 0] = 0. # No negative values in reference spectrum
                    corr, residual = compute_and_solve_model(
                        wn, app[s], ref=ref,
                        modelparams=modelparams, solvingparams=solvingparams,
                        plot=(iteration, iterations))
                    # print("pixel %5d: iter %3d  residual %7.3g" %
                    #       (s, iteration, residual))
                    if autoiterations:
                        if residual < residuals[s]:
                            corrected[s, :] = corr
                            unimproved =  unimproved + 1 if \
                                residual > residuals[s] * targetrelresiduals else 0
                            residuals[s] = residual
                        else:
                            unimproved += autoupadd
                        if unimproved > automax:
                            break
                    ref = corr
                if not autoiterations:
                    corrected[s, :] = corr

                if verbose:
                    print("pixel %5d: iter %3d  residual %7.3g  time %f" %
                          (s, iteration, residual, monotonic()-startt))
                if progressCallback:
                    progressA += 1
                    progressCallback(progressA, progressB)
                if progressPlotCallback and len(app) < 50:
                    progressPlotCallback(ref, (s + 1, len(app) + 1))

    return corrected.squeeze() if squeeze else corrected





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat May 15 19:58:51 2021

@author: carl
"""

import json

class Parameters:
    """
    A class representing all the settings that can be made in some UI,
    saved/loaded and e.g. used to start a batch job.
    """
    def __init__(self, items=None):
        if items is not None:
            self.__dict__.update(items)

    def save(self, filename):
        with open(filename, 'w') as fp:
            json.dump(vars(self), fp, indent=4)

    def load(self, filename):
        with open(filename, 'r') as fp:
            data = json.load(fp)
            self.__dict__.update(data)

    def filtered(self, prefix):
        "Returns a dict of parameters that start with the given prexix"
        return { k: v for k, v in vars(self).items() if k.startswith(prefix)}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  4 16:21:46 2020

@author: carl
"""

import numpy as np
from scipy.signal import savgol_filter, tukey

from . import baseline
from .util import load_reference, find_wn_ranges


def cut_wn(wn, y, ranges):
    """
    Cut a set of spectra, leaving only the given wavenumber range(s).

    Parameters
    ----------
    wn: array of wavenumbers, sorted in either direction
    y: array of spectra, shape (..., wavenumber)
    ranges: list or numpy array of shape (..., 2) with desired wavenumber
    ranges in pairs (low, high)
    Returns: (wavenumbers, spectra) with data in the given wavenumber ranges
    """
    if isinstance(ranges, list):
        ranges = np.array(ranges)
    inrange = lambda w: ((w >= ranges[...,0]) & (w <= ranges[...,1])).any()
    ix = np.array([inrange(w) for w in wn])
    return wn[ix], y[...,ix]

def atmospheric(wn, y, atm=None, cut_co2 = True, extra_iters=5,
                extra_factor=0.25, smooth_win=9, progressCallback = None):
    """
    Correct for atmospheric gases.

    Apply atmospheric correction to multiple spectra, subtracting as much
    of the atompsheric spectrum as needed to minimize the sum of squares of
    differences between consecutive points in the corrected spectra.
    Each supplied range of wavenumbers is corrected separately.

    Parameters
    ----------
        wn: array of wavenumbers, sorted in either direction
        y: array of spectra in the order (pixel, wavenumber), or just one spectrum
        atm: atmospheric spectrum; if None, load the default
        cut_co2: replace the CO2 region with a neatly fitted spline
        extra_iters: number of iterations of subtraction of a locally reshaped atmospheric spectrum
        (needed if the relative peak intensities are not always as in the atmospheric reference)
        extra_factor: how much of the reshaped atmospheric spectrum to remove per iteration
        smooth_win: window size (in cm-1) for smoothing of the spectrum in the atm regions
        progressCallback(int a, int b): callback function called to indicated that the processing
        is complete to a fraction a/b.

    Returns
    -------
        Tuple of (spectra after correction, array of correction factors;
                  shape (spectra,ranges))
    """
    squeeze = False
    yorig = y
    if y.ndim == 1:
        y = y[None,:]
        squeeze = True
    else:
        y = y.copy()

    if atm is None or (isinstance(atm, str) and atm == ''):
        atm = load_reference(wn, what='water')
    elif isinstance(atm, str):
        atm = load_reference(wn, matfilename=atm)
    else:
        atm = atm.copy()

    # ranges: numpy array (n, 2) of n non-overlapping wavenumber ranges
    # (typically for H2O only), or None
    # extra_winwidth: width of the window (in cm-1) used to locally
    # reshape the atm spectrum
    ranges = [[399, 776], [1300, 2100], [3410, 3960], [2190, 2480]]
    do_ranges = np.ones(len(ranges), dtype=bool)
    n_gas = np.array([0, 0, 0, 1])
    extra_winwidth = [100, 30, 150, 40]
    if cut_co2:
        do_ranges[n_gas == 1] = False

    if ranges is None:
        ranges = np.array([0, len(wn)])
    else:
        ranges = find_wn_ranges(wn, ranges)

    for i, (p, q) in enumerate(ranges):
        if not do_ranges[i]:
            continue
        if q - p < 2:
            do_ranges[i] = False
            continue
        # atm[p:q] -= baseline.straight(wn[p:q], atm[p:q])
        if np.abs(atm[p:q]).sum() == 0:
            do_ranges[i] = False

    savgolwin = 1 + 2 * int(smooth_win * (len(wn) - 1) / np.abs(wn[0] - wn[-1]))

    corr_ranges = do_ranges.sum()
    if progressCallback:
        progressA = 0
        progressB = 1 + corr_ranges * (extra_iters + (1 if savgolwin > 1 else 0))
        progressCallback(progressA, progressB)

    dh = atm[:-1] - atm[1:]
    dy = y[:,:-1] - y[:,1:]
    dh2 = np.cumsum(dh * dh)
    dhdy = np.cumsum(dy * dh, 1)
    for i, (p, q) in enumerate(ranges[do_ranges]):
        r = q-2 if q <= len(wn) else q-1
        az = ((dhdy[:,r] - dhdy[:,p-1]) / (dh2[r] - dh2[p-1])
                    ) if p > 0 else (dhdy[:,r] / dh2[r])
        y[:, p:q] -= az[..., None] @ atm[None, p:q]

    if progressCallback:
        progressA += 1
        progressCallback(progressA, progressB)

    for pss in range(extra_iters):
        for i, (p, q) in enumerate(ranges[do_ranges]):
            window = 2 * int(extra_winwidth[i] * (len(wn) - 1) / np.abs(wn[0] - wn[-1]))
            winh = (window+1)//2
            dy = y[:,:-1] - y[:,1:]
            dhdy = np.cumsum(dy * dh, 1)
            aa = np.zeros_like(y)
            aa[:,1:winh+1] = dhdy[:,1:window:2] / np.maximum(dh2[1:window:2], 1e-8)
            aa[:,1+winh:-winh-1] = (dhdy[:,window:-1] -
              dhdy[:,:-1-window]) / np.maximum(dh2[window:-1] - dh2[:-1-window], 1e-8)
            aa[:,-winh-1:-1] = (dhdy[:,-1:] -
              dhdy[:,-1-window:-1:2]) / np.maximum(dh2[-1] - dh2[-1-window:-1:2], 1e-8)
            aa[:, 0] = aa[:, 1]
            aa[:, -1] = aa[:, -2]
            aa = savgol_filter(aa, window + 1, 3, axis=1)
            y[:, p:q] -= extra_factor * aa[:, p:q] * atm[p:q]
            if progressCallback:
                progressA += 1
                progressCallback(progressA, progressB)

    if savgolwin > 1:
        for i, (p, q) in enumerate(ranges[do_ranges]):
            if q - p < savgolwin: continue
            y[:, p:q] = savgol_filter(y[:, p:q], savgolwin, 3, axis=1)
            if progressCallback:
                progressA += 1
                progressCallback(progressA, progressB)

    if cut_co2:
        rng = np.array([[2190, 2260], [2410, 2480]])
        rngm = rng.mean(1)
        rngd = rngm[1] - rngm[0]
        cr = find_wn_ranges(wn, rng).flatten()
        if cr[1] - cr[0] > 2 and cr[3] - cr[2] > 2:
            a = np.empty((4, len(y)))
            a[0:2,:] = np.polyfit((wn[cr[0]:cr[1]]-rngm[0])/rngd, y[:,cr[0]:cr[1]].T, deg=1)
            a[2:4,:] = np.polyfit((wn[cr[2]:cr[3]]-rngm[1])/rngd, y[:,cr[2]:cr[3]].T, deg=1)
            P,Q = find_wn_ranges(wn, rngm[None,:])[0]
            t = np.interp(wn[P:Q], wn[[Q,P] if wn[0] > wn[-1] else [P,Q]], [1, 0])
            tt = np.array([-t**3+t**2, -2*t**3+3*t**2, -t**3+2*t**2-t, 2*t**3-3*t**2+1])
            pt = a.T @ tt
            y[:, P:Q] += (pt - y[:, P:Q]) * tukey(len(t), .3)

    corrs = np.zeros(2)
    ncorrs = np.zeros_like(corrs)
    if cut_co2:
        do_ranges[n_gas == 1] = True
    for i, (p, q) in enumerate(ranges[do_ranges]):
        if q > p:
            corr = np.abs(yorig[:, p:q] - y[:, p:q]).sum(1) / np.maximum(
                np.abs(yorig[:, p:q]), np.abs(y[:, p:q])).sum(1)
            g = n_gas[do_ranges][i]
            corrs[g] += corr.mean()
            ncorrs[g] += 1
    corrs[ncorrs > 1] = corrs[ncorrs > 1] / ncorrs[ncorrs > 1]

    return (y.squeeze() if squeeze else y), corrs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 14 16:04:24 2019

@author: carl

A QLineEdit widget that holds a double and handles some validation but
doesn't require it to be valid at all points during editing

"""

from PyQt5.QtWidgets import QLineEdit
#from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtCore import pyqtSlot #, QLocale

class LineDouble(QLineEdit):
    def __init__(self, parent=None, vmin=0, vmax=100, fmt='%.2f'):
        QLineEdit.__init__(self)
        self.default = (vmin + vmax) / 2
        self.min = vmin
        self.max = vmax
        self.setFormat(fmt)
        self.editingFinished.connect(self.validate)
#        self.setValidator(QDoubleValidator(min, max, 1000, parent=self))

#    @pyqtSlot(float, float)
#    def setRange(self, min, max):
#        self.min = min
#        self.max = max
#        self.validator().setRange(min, max, 1000)

    def setRange(self, vmin, vmax, default=None):
        assert(vmin <= vmax)
        self.min = vmin
        self.max = vmax
        if default is not None:
            self.default = default
        elif self.text() != '':
            self.default = float(self.text())
        else :
            self.default = (vmin+vmax)/2.
        assert(vmin <= self.default <= vmax)

    def setFormat(self, fmt):
        self.fmt=fmt

    @pyqtSlot()
    def validate(self):
        self.setValue(self.value())

    @pyqtSlot(float)
    def setValue(self, val=None):
#        if sci:
#            self.setText(QLocale().toString(val, 'g', 3))
#        else:
#            self.setText(QLocale().toString(val, 'f', 2))
        if val is None:
            val = self.default
        self.setText(self.fmt % min(max(val, self.min), self.max))

    def hasAcceptableValue(self):
        try:
            float(self.text())
        except Exception:
            return False
        return True

    def value(self):
#        f = QLocale().toFloat(self.text())[0]
#        if self.hasAcceptableInput():
#            return f
#        return max(self.validator().bottom(), min(self.validator().top(), f))
        v = self.default
        try:
            v = float(self.text())
        except Exception:
            pass
        return min(max(v, self.min), self.max)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 27 00:23:22 2021

@author: carl
"""

import numpy as np
import math

def impute_from_neighbors(data, spatial=True, spectral=False):
    """
    Impute missing intensities as mean of surrounding values in the image.

    The mean of nearest neighbor pixels replaces all NaNs that have at least
    one non-NaN neighbor, and this is iterated until all NaNs are replaced.

    Parameters
    ----------
    data : numpy.ndarray
        Intensities stored as data[x, y, ..., wavenumber].
    spatial : bool
        Use spatial nearest neighbors (n-1 first dimensions).
    spectral : bool
        Use spectral nearest neighbors (last dimension).

    Returns
    -------
    data : numpy.ndarray
        Transformed data.

    """
    todo = np.argwhere(~np.isfinite(data))
    if not len(todo):
        return data
    if len(todo) == data.size:
        raise ValueError('Imputation requires at least one non-NaN in input')
    wh = data.shape
    data = data.copy()
    if spatial and spectral:
        sumdims = range(len(wh))
    elif spatial:
        sumdims = range(len(wh) - 1)
    elif spectral:
        sumdims = [-1]
    else:
        raise ValueError('Imputation must use spatial and/or spectral neighbors')

    while len(todo):
        # print('todo:', len(todo))
        done = []
        donev = []
        stilltodo = []
        for xy in todo:
            nsum = 0.
            nn = 0
            for d in sumdims:
                c = xy[d]
                if c > 0:
                    xy[d] = c - 1
                    v = data[tuple(xy)]
                    if math.isfinite(v):
                        nsum = nsum + v
                        nn = nn + 1
                if c + 1 < wh[d]:
                    xy[d] = c + 1
                    v = data[tuple(xy)]
                    if math.isfinite(v):
                        nsum = nsum + v
                        nn = nn + 1
                xy[d] = c
            if nn:
                done.append(tuple(xy))
                donev.append(nsum / nn)
            else:
                stilltodo.append(xy)
        if not len(done):
            # Should perhaps raise an exception instead
            break
        data[tuple(np.asarray(done).T)] = donev
        todo = stilltodo
    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 28 01:07:18 2019

@author: carl
"""

from io import BytesIO
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib, cycler
import matplotlib.pyplot as plt
from collections import OrderedDict
import numpy as np
import random
from sklearn.cluster import KMeans
from sklearn.metrics import pairwise_distances_argmin_min


# Silly naming covention in this class: camelCase for public functions,
# lower_case for private stuff
class ProjectionWidget(FigureCanvas):
    changedSelected = pyqtSignal(int)
    updatedProjection = pyqtSignal(np.ndarray)

    def __init__(self, parent=None):
        self.fig = Figure()
        self.ax = self.fig.add_subplot(111)
        self.ax.set_axis_off()
        self.ax.set_position([0,0,1,1])
        FigureCanvas.__init__(self, self.fig)

        self.wavenumber = None
        self.raw = None
        self.wavenumber = None  # Wavenumbers, used for area-under-curve
        self.wh = None          # Image shape for non-pixelxy
        self.pixelxy = None     # (x,y) for all pixels or None

        self.spatial = False    # Display in pixelxy-mode
        self.clustered = False  # Select based on cluster centers
        self.selected = OrderedDict() # keys indicate selected pixels
        self.pixel_levels = None # Current intensity of each pixel
        self.cmap = plt.get_cmap().name # For img/fill
        self.fillmode = False   # Fill the points
        self.plot_cmap = None   # Name of cmap for borders, if any
        self.patches = {}       # Patch objects indexed by pixel number
        self.pop_patches = {}   # Ditto in pop-out plot
        # self.white_patches = {} # Ditto in white light subplot
        self.rect_size = 2      # Size of rectancles in µm
        self.rect_color = 'b'   # Default patch edge color
        self.pixel_visible = None # None, or True for pixels within image
        self.image = None       # ..io.Image object
        # Main/background image with placeholder data
        self.bgimg = self.ax.imshow(
            np.random.rand(16, 16), cmap=self.cmap, zorder=0, aspect='equal',
            origin='lower')
        self.pop_bgimg = None   # Ditto in pop-out plot
        self.pop_fig = None     # The pop-out Figure
        self.pop_ax = None      # ...and its Axes
        self.pop_cb = None      # ...and its Colorbar
        self.mpl_connect('button_press_event', self.onclick)

    def make_spatial_patches(self, patches, ax):
        rs = self.rect_size
        for xy in self.pixelxy:
            rect = matplotlib.patches.Rectangle(
                (xy[0] - rs/2, xy[1] - rs/2), rs, rs,
                fill=True, color=self.rect_color)
            ax.add_patch(rect)
            patches[len(patches)] = rect

    def setData(self, wn, raw, pixelxy=None):
        "Set spectral data and positions and clear selection"
        self.wavenumber = wn
        self.raw = raw
        self.pixelxy = pixelxy
        # New data implies cleared selection of pixels
        if self.spatial:
            for patches in [self.patches, self.pop_patches]:
                for p in patches.values():
                    p.remove()
            self.patches = {}
            self.pop_patches = {}
            self.selected = OrderedDict()
        else:
            self.clear_selected()
        self.spatial = pixelxy is not None
        if self.spatial:
            self.make_spatial_patches(self.patches, self.ax)
            if self.pop_fig:
                self.make_spatial_patches(self.pop_patches, self.pop_ax)
        self.pixel_visible = None

    def setDimensions(self, wh):
        "Set the shape of the image, with override if too unsquare"
        p = wh[0] * wh[1]
        if p <= 0:
            return
        if wh[0] > 8 * wh[1] or wh[1] > 8 * wh[0]:
            h = int(np.sqrt(p))
            self.wh = ((p + h - 1) // h, h)
        else:
            self.wh = wh
        self.trigger_redraw()

    def bgimglist(self):
        return [self.bgimg, self.pop_bgimg] if self.pop_bgimg \
            else [self.bgimg]

    def trigger_redraw(self):
        self.draw_idle()
        if self.pop_fig:
            self.pop_fig.canvas.draw_idle()


    def setImage(self, image):
        "Set the background or white light image (or None to clear it)"
        if image is None:
            for bgimg in self.bgimglist():
                bgimg.set_data([[[200,200,200,255]]])
                if self.spatial:
                    minxy = np.min(self.pixelxy, axis=0)
                    maxxy = np.max(self.pixelxy, axis=0)
                    dxy = (maxxy - minxy) * .2
                    minxy = minxy - dxy
                    maxxy = maxxy + dxy
                    bgimg.set_extent((minxy[0], maxxy[0], minxy[1], maxxy[1]))
                    self.pixel_visible = [True] * len(self.pixelxy)
                    self.update_patch_fill()
                else:
                    bgimg.set_extent((0, self.wh[0], 0, self.wh[1]))
            self.trigger_redraw()
            return

        if image.fmt is not None:
            imgdata = plt.imread(BytesIO(image.data), format=image.fmt)
        elif image.data is not None:
            imgdata = image.data

        if self.spatial:
            hwh = np.array(image.wh) / 2
            for bgimg in self.bgimglist():
                bgimg.set_data(imgdata)
                bgimg.set_extent((image.xy[0] - hwh[0], image.xy[0] + hwh[0],
                                  image.xy[1] - hwh[1], image.xy[1] + hwh[1]))
                bgimg.autoscale()
            self.pixel_visible = [
                (np.abs(image.xy - xy) < (hwh + self.rect_size / 2)).all()
                for xy in self.pixelxy ]
            self.update_patch_fill()
        self.trigger_redraw()

    def setRectSize(self, size):
        if size <= 0:
            raise ValueError("Rectangle size must be positive")
        dpos = (self.rect_size - size) / 2
        self.rect_size = size
        for patches in [self.patches, self.pop_patches]:
            for p in patches.values():
                x, y = p.get_xy()
                p.set_bounds(x + dpos, y + dpos, size, size)
        self.trigger_redraw()

    def setProjection(self, method, wavenumix):
        "Update pixel colors. Methods 0=area, 1=max, 2=wavenum"
        if method == 0:
            assert self.wavenumber is not None
            data = np.trapz(self.raw, self.wavenumber, axis=1)
            if not np.isfinite(data.sum()):
                data = np.trapz(np.nan_to_num(self.raw), self.wavenumber, axis=1)
            if self.wavenumber[0] > self.wavenumber[-1]:
                data = -data
        elif method == 1:
            data = np.nanmax(self.raw, axis=1)
        elif method == 2:
            data = self.raw[:, wavenumix]
        self.pixel_levels = data

        if self.spatial:
            self.update_patch_fill()
        else:
            px = self.wh[0] * self.wh[1]
            if px != len(data):
                d2 = np.empty((px))
                d2[:len(data)] = data
                d2[len(data):] = data.min()
                data = d2

            data = data.reshape(self.wh[::-1])
            for bgimg in self.bgimglist():
                bgimg.set_data(data)
                bgimg.set_extent((0, self.wh[0], 0, self.wh[1]))
                bgimg.autoscale()

        self.trigger_redraw()
        self.updatedProjection.emit(self.pixel_levels)

    def getWavenumbers(self):
        return self.wavenumber

    def visible_levels(self):
        "Get min and max visible pixel levels, for color range"
        vislevels = self.pixel_levels if self.pixel_visible is None \
            else self.pixel_levels[self.pixel_visible]
        return vislevels.min(), vislevels.max()

    def update_patch_fill(self):
        "Update fill color of pixelxy patches, normalized to visible range"
        if not self.spatial:
            return
        vislevels = self.visible_levels()
        colors = matplotlib.cm.ScalarMappable(
            norm=matplotlib.colors.Normalize(
                vmin=vislevels[0], vmax=vislevels[1], clip=True),
            cmap=self.cmap).to_rgba(self.pixel_levels)
        for pits in [self.patches, self.pop_patches]:
            for n, p in pits.items():
                if self.fillmode and n in self.selected:
                    p.set_facecolor(p.get_edgecolor())
                else:
                    p.set_facecolor(colors[n])

    def setGlobalColorCycle(self, name):
        """
        Set the color used for line plots globally

        Parameters
        ----------
        name : str
            Name of cmap or "all [color]" or None to leave it unchanged.

        Returns
        -------
        cols : list
            List of colors to cycle through, or None if unchanged
        """
        if name is None:
            if self.plot_cmap is None:
                return plt.rcParams['axes.prop_cycle'].by_key()['color']
            name = self.plot_cmap
        self.plot_cmap = None
        if name[:4] == 'all ':
            cols = [matplotlib.colors.to_rgb(name[4:])]
        else:
            cm = plt.get_cmap(name)
            if hasattr(cm, 'colors') and cm.N < 32:
                cols = cm.colors
            else:
                n = max(len(self.selected), 1)
                d = .1 if name == 'gray' else 0
                cols = [cm(x) for x in np.linspace(d, 1-d, n)]
                self.plot_cmap = name
        plt.rcParams['axes.prop_cycle'] = cycler.cycler(
            'color', cols)
        return cols

    @pyqtSlot('QString')
    def setCmap(self, s):
        "Update colors of background image"
        self.cmap = s
        self.update_patch_fill()
        for bgimg in self.bgimglist():
            bgimg.set_cmap(s)
        self.trigger_redraw()

    def setFill(self, fill):
        "Fill/unfill pixelxy patches"
        self.fillmode = fill
        # Spatial-mode pixels are always filled (with level or edgecolor)
        if self.spatial:
            self.update_patch_fill()
        else:
            for p in self.patches.values():
                p.set_fill(fill)
            for p in self.pop_patches.values():
                p.set_fill(fill)
        self.trigger_redraw()

    def updatePlotColors(self, name=None):
        colors = self.setGlobalColorCycle(name)
        i = 0
        for s in self.selected.keys():
            c = colors[i % len(colors)]
            self.patches[s].set_edgecolor(c)
            if s in self.pop_patches:
                self.pop_patches[s].set_edgecolor(c)
            i = i + 1
        self.trigger_redraw()

    def getSelected(self):
        return list(self.selected.keys())

    def getSelectedData(self):
        if self.raw is None:
            return None
        return self.raw[self.getSelected(), :]

    def clear_selected(self):
        if self.spatial:
            for s in self.selected:
                self.patches[s].set_edgecolor(self.rect_color)
                if s in self.pop_patches:
                    self.pop_patches[s].set_edgecolor(self.rect_color)
        else:
            for p in self.patches.values():
                p.remove()
            self.patches = {}
            for p in self.pop_patches.values():
                p.remove()
            self.pop_patches = {}
        self.selected = OrderedDict()

    def new_patch(self, p, colorix):
        "Construct Patch for pixel p in non-spatial mode"
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        return matplotlib.patches.Rectangle(
            (p % self.wh[0], p // self.wh[0]), .999, .999, fill=self.fillmode,
            color=colors[colorix % len(colors)])

    def color_patch(self, p, color):
        "Set edge color of patch(es) for pixel p"
        for pps in [self.patches, self.pop_patches]:
            if p in pps:
                pps[p].set_edgecolor(color)
                if self.fillmode:
                    pps[p].set_facecolor(color)

    def select_point(self, p):
        assert p not in self.selected
        if self.spatial:
            colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
            c = colors[len(self.selected) % len(colors)]
            self.color_patch(p, c)
            self.selected[p] = True
        else:
            if p < 0 or p >= len(self.raw):
                return False
            pnum = len(self.selected)
            pat = self.new_patch(p, pnum)
            self.patches[p] = pat
            self.ax.add_patch(pat)
            self.selected[p] = True
            if self.pop_ax is not None:
                pat = self.new_patch(p, pnum)
                self.pop_patches[p] = pat
                self.pop_ax.add_patch(pat)
        return True

    def deselect_point(self, p):
        del self.selected[p]
        if self.spatial:
            self.color_patch(p, self.rect_color)
        else:
            self.patches.pop(p).remove()
            if p in self.pop_patches:
                self.pop_patches.pop(p).remove()

    def setSelectedCount(self, n, clustered):
        if self.raw is None:
            return
        if n > len(self.raw):
            n = len(self.raw)
        if n == len(self.selected) and clustered == self.clustered:
            return
        self.clustered = clustered
        if n == len(self.raw):
            for p in range(len(self.raw)):
                if p not in self.selected:
                    self.select_point(p)
        elif clustered and n > 0:
            # (not wasc or n != len(self.selected)) and
            km = KMeans(n_clusters=min(n, len(self.raw)), n_init=1)
            km.fit(self.raw)
            closest = pairwise_distances_argmin_min(
                km.cluster_centers_, self.raw)[0]
            self.clear_selected()
            for p in closest:
                self.select_point(p)
        if n < len(self.selected):
            for p in list(self.selected.keys())[n:]:
                self.deselect_point(p)
        while len(self.selected) < n:
            p = random.randrange(len(self.raw))
            if p not in self.selected:
                self.select_point(p)
        self.updatePlotColors()
        self.trigger_redraw()
        self.changedSelected.emit(len(self.selected))

    def onclick(self, event):
        if event.xdata is None:
            return
        if event.inaxes == self.ax:
            patches = self.patches
        elif event.inaxes == self.pop_ax:
            patches = self.pop_patches
        else:
            return
        changed = False  # Anything selected/deselected?
        if self.spatial:
            for p, pat in patches.items():
                if pat.contains(event)[0]:
                    changed = True
                    if p in self.selected:
                        self.deselect_point(p)
                    else:
                        self.select_point(p)
        elif self.wh is not None:
            x, y = (int(event.xdata), int(event.ydata))
            if not (0 <= x < self.wh[0] and 0 <= y < self.wh[1]):
                return
            p = x + y * self.wh[0]
            if p in self.selected:
                self.deselect_point(p)
                changed = True
            elif self.select_point(p):
                changed = True
        if changed:
            self.updatePlotColors()
            self.trigger_redraw()
            self.changedSelected.emit(len(self.selected))

    def onclose(self, event):
        "Pop-out window was close; stop drawing there"
        self.pop_fig = None
        self.pop_ax = None
        self.pop_bgimg = None
        self.pop_cb = None
        self.pop_patches = {}

    def draw(self):
        FigureCanvas.draw(self)
        # if plt.fignum_exists(self.objectName()):
        #     self.popOut()

    def popOut(self):
        fig = plt.figure(self.objectName(), tight_layout=dict(pad=.6))
        if self.pop_fig != fig:
            self.pop_fig = fig
            fig.canvas.mpl_connect('button_press_event', self.onclick)
            fig.canvas.mpl_connect('close_event', self.onclose)
            self.pop_ax = fig.gca()
            self.pop_bgimg = self.pop_ax.imshow(
                self.bgimg.get_array(), cmap=self.cmap,
                zorder=0, aspect='equal', extent=self.bgimg.get_extent(),
                origin='lower')
            self.pop_bgimg.autoscale()

            self.pop_patches = {}
            if self.spatial:
                self.make_spatial_patches(self.pop_patches, self.pop_ax)
                self.update_patch_fill()
            else:
                self.pop_cb = fig.colorbar(self.pop_bgimg, ax=self.pop_ax)
                for p in self.selected:
                    pat = self.new_patch(p, len(self.pop_patches))
                    self.pop_patches[p] = pat
                    self.pop_ax.add_patch(pat)

            fig.show()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 27 22:55:02 2019

@author: carl
"""

import traceback
import numpy as np
import scipy.signal, scipy.io
# from scipy.interpolate import interp1d
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot

from octavvs.io import DecompositionData, Parameters
from octavvs.algorithms import decomposition, imputation
import os.path
import time
import sklearn.cluster


class DecompWorker(QObject):
    """
    Worker thread class for the heavy parts of the decomposition
    """
    # Signals for when finished or failed, and for progress indication
    # done: iteration, spectra, concentrations, errors
    done = pyqtSignal(np.ndarray, np.ndarray, np.ndarray)
    stopped = pyqtSignal()
    failed = pyqtSignal(str, str)
    progress = pyqtSignal(int, float) # iteration, rel. error
    # done: iteration, spectra, concentrations, errors
    progressPlot = pyqtSignal(int, np.ndarray, np.ndarray, list)

    fileLoaded = pyqtSignal(int)
    loadFailed = pyqtSignal(str, str, str) # file, msg, details

    batchProgress = pyqtSignal(int, int)
    batchDone = pyqtSignal(bool)

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self.halt = False

    @pyqtSlot(int)
    def loadFile(self, data, num):
        "Load file number num in the data object, emitting a signal on failure"
        try:
            file = data.filenames[num]
            if file == data.curFile:
                return True
            data.read_matrix(file)
        except (RuntimeError, FileNotFoundError) as e:
            self.loadFailed.emit(file, str(e), '', False)
        except Exception as e:
            self.loadFailed.emit(file, str(e), traceback.format_exc())
        else:
            self.fileLoaded.emit(num)
            return True
        return False

    def emitProgress(self, *pargs):
        """Signal progress and check for user interruption."""
        if self.halt:
            raise InterruptedError('interrupted by user')
        self.progress.emit(*pargs)


    def callDecomp(self, data, params):
        """
        Run decomposition stuff.
        """
        y = data.raw
        if params.dcImpute:
            sh = [data.wh[1], data.wh[0], y.shape[-1]]
            y = imputation.impute_from_neighbors(y.reshape(sh))
            y = y.reshape([-1, sh[-1]])
        if data.decomposition_roi is not None:
            y = y[data.decomposition_roi, :]

        okwn = np.isfinite(y.sum(axis=0))
        if okwn.sum() == len(y):
            okwn = None
        else:
            y = y[:, okwn]

        # Restore hidden wavenumbers (replace with 0)
        def restore_hidden_wn(spectra):
            if okwn is not None:
                stmp = spectra
                spectra = np.zeros((len(spectra), len(okwn)))
                spectra[:, okwn] = stmp
            return spectra

        c_first = params.dcStartingPoint != 0
        nonneg = [True, True]

        if params.dcDerivative:
            y = scipy.signal.savgol_filter(
                y, window_length=params.dcDerivativeWindow,
                polyorder=params.dcDerivativePoly,
                deriv=params.dcDerivative, axis=1)
            nonneg[int(c_first)] = False

        base_error = np.linalg.norm(y - y.mean(0))**2
        if c_first:
            y = y.T

        if params.dcInitialValues == 'simplisma':
            # simplisma_noise = 0.1
            initst = decomposition.simplisma(
                y.T, params.dcComponents, params.dcInitialValuesNoise)[0]
        elif params.dcInitialValues == 'kmeans':
            km = sklearn.cluster.MiniBatchKMeans(
                n_clusters=params.dcComponents).fit(y)
            initst = km.cluster_centers_
        elif params.dcInitialValues == 'clustersubtract':
            initst = decomposition.clustersubtract(
                y, components=params.dcComponents,
                skewness=params.dcInitialValuesSkew,
                power=params.dcInitialValuesPower)
        else:
            raise ValueError('Unknown params.dcInitialValues')

        if c_first and params.dcDerivative:
            # If concentrations-first is combined with derivatives,
            # the initial concentrations need to be made positive.
            initst = np.abs(initst)

        if params.dcAlgorithm == 'mcr-als':
            acceleration = None
        elif params.dcAlgorithm == 'mcr-als-anderson':
            acceleration = 'Anderson'
        else:
            raise ValueError('Unknown params.dcAlgorithm')

        self.emitProgress(0, 0.)
        if c_first:
            data.add_decomposition_data(0, None, initst)
            self.progressPlot.emit(0, np.array(()), initst, [])
        else:
            rist = restore_hidden_wn(initst)
            data.add_decomposition_data(0, rist, None)
            self.progressPlot.emit(0, rist, np.array(()), [])

        update_interval = 3
        def cb_iter(it, errs, spectra, concentrations):
            self.emitProgress(len(errs), errs[-1] / base_error)
            t = time.monotonic()
            if t > cb_iter.iter_next:
                cb_iter.iter_next = t + update_interval
                if c_first:
                    concentrations, spectra = (spectra, concentrations)
                spectra = restore_hidden_wn(spectra)
                self.progressPlot.emit(it + 1, spectra, concentrations, errs)
        cb_iter.iter_next = time.monotonic()

        cweight = None
        if params.dcContrast:
            cweight = ('B' if params.dcContrastConcentrations == c_first \
                       else 'A', params.dcContrastWeight)

        # tt = time.monotonic()
        spectra, concentrations, errors = decomposition.mcr_als(
            y, initst, maxiters=params.dcIterations,
            nonnegative=nonneg,
            # tol_abs_error=params.dcTolerance * base_error,
            tol_rel_improv=params.dcTolerance * .01,
            tol_iters_after_best=40, callback=cb_iter,
            acceleration=acceleration, normalize='B' if c_first else 'A',
            contrast_weight=cweight)
        if c_first:
            concentrations, spectra = (spectra, concentrations)
        # print('Time:', time.monotonic()-tt)

        spectra = restore_hidden_wn(spectra)

        errors = np.asarray(errors)
        data.add_decomposition_data(len(errors), spectra, concentrations)
        data.set_decomposition_errors(errors)
        self.done.emit(spectra, concentrations, errors)
        return True


    @pyqtSlot(DecompositionData, Parameters)
    def decompose(self, data, params):
        """ Run decomposition on all or a subset of the raw data.
        Parameters:
            data: DecompositionData object with raw data
            params: Parameters with things set
        """
        try:
            self.callDecomp(data, params)
        except InterruptedError:
            self.stopped.emit()
        except Exception as e:
            traceback.print_exc()
            self.failed.emit(repr(e), traceback.format_exc())
        self.stopped.emit()



    @pyqtSlot(DecompositionData, Parameters)
    def startBatch(self, data, params):
        """
        Run the batch processing of all the files listed in 'data'
        Parameters:
            data: DecompositionData object with one or more files
            params: Parameters object from the user
        """
        try:
            for fi in range(len(data.filenames)):
                self.batchProgress.emit(fi, len(data.filenames))
                # rcd_dir = params.directory if params.dirMode else None
                if not self.loadFile(data, fi):
                    continue

                data.set_decomposition_settings(params)
                self.callDecomp(data, params)
                fn = os.path.splitext(os.path.basename(data.curFile))[0]
                filename = os.path.join(params.dcDirectory, fn) + '.dcn'
                data.save_rdc(filename=filename)

            self.batchDone.emit(True)
            return

        except InterruptedError:
            self.stopped.emit()
        except Exception as e:
            traceback.print_exc()
            self.failed.emit(repr(e), traceback.format_exc())
        self.batchDone.emit(False)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan  9 16:16:44 2020

@author: carl
"""

import numpy as np
# from .image import Image

class OmnicReader:
    """
    A simple wrapper around the spectrochempy loader for OMNIC files, with
    very limited ability to convert data into the shape assumed by OCTAVVS.

    Member variables:
        wh: image dimensions (w, h)
        xy: points coordinates [(x,y)]
        wavenum: array of wavenumbers
        AB: data matrix, array(points, len(wn))
        image: tuple (img_data_bytes, img_type_str) (unused)
    """

    def __init__(self, filename=None):
        """
        filename:  spa/spg file to load data from
        """
        self.wh = None
        self.xy = None
        self.wavenum = None
        self.AB = None
        self.images = []
        if filename is not None:
            self.load(filename)

    def load(self, filename):
        try:
            import spectrochempy as scp
        except ImportError:
            raise RuntimeError(
                'Importing an OMNIC file requires the spectrochempy package. '
                'E.g. "conda install -c conda-forge -c cantera -c spectrocat '
                'spectrochempy" or "pip install spectrochempy"')
        om = scp.read_omnic(filename)
        self.AB = om.data
        self.wavenum = om.x.values.m
        assert self.AB.ndim == 2 and (self.AB.shape[1],) == self.wavenum.shape





import os
import os.path
import numpy as np
#import fnmatch
#import traceback
from pkg_resources import resource_filename

#from PyQt5.QtCore import QThread, pyqtSignal, pyqtSlot, Qt
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import pyqtSignal
from . import constants, uitools
from octavvs.ui import NoRepeatStyle
# from ..io import Image
from scipy.spatial import KDTree

ImageVisualizerUi = uic.loadUiType(
    resource_filename(__name__, "imagevisualizer.ui"))[0]

class ImageVisualizerWidget(QWidget, ImageVisualizerUi):
    """
    Widget with a bunch of graphical components
    """
    updatePixelSize = pyqtSignal(float)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)
        self.spinBoxSpectra.setStyle(NoRepeatStyle())
        self.lineEditWavenumber.setFormat("%.2f")

        self.pushButtonExpandProjection.clicked.connect(self.plot_raw.popOut)
        self.comboBoxCmaps.currentTextChanged.connect(self.plot_raw.setCmap)
        self.comboBoxMethod.currentIndexChanged.connect(
            self.updateSpectralImage)
        self.checkBoxPixelFill.toggled.connect(self.plot_raw.setFill)
        self.lineEditPixelSize.setRange(1e-2, 1e3)
        self.lineEditPixelSize.setFormat('%g')
        self.lineEditPixelSize.editingFinished.connect(self.pixelSizeEdit)
        self.horizontalSliderWavenumber.valueChanged.connect(
            self.wavenumberSlide)
        self.lineEditWavenumber.editingFinished.connect(self.wavenumberEdit)
        self.comboBoxPlotCmap.currentTextChanged.connect(
            self.plot_raw.updatePlotColors)

        self.updatePixelSize.connect(self.plot_raw.setRectSize)
        self.plot_raw.changedSelected.connect(self.selectedSpectraUpdated)
        self.plot_spectra.clicked.connect(self.plot_spectra.popOut)
        self.spinBoxSpectra.valueChanged.connect(self.selectSpectra)
        self.checkBoxAutopick.toggled.connect(self.selectSpectra)

    def pixelSizeEdit(self):
        self.updatePixelSize.emit(self.lineEditPixelSize.value())

    def wavenumberEdit(self):
        uitools.box_to_slider(
                self.horizontalSliderWavenumber, self.lineEditWavenumber,
                self.plot_raw.getWavenumbers(),
                uitools.ixfinder_nearest)

    def wavenumberSlide(self):
        uitools.slider_to_box(
                self.horizontalSliderWavenumber, self.lineEditWavenumber,
                self.plot_raw.getWavenumbers(),
                uitools.ixfinder_nearest)
        self.updateSpectralImage()

    def updateSpectralImage(self):
        meth = self.comboBoxMethod.currentIndex()
        iswn = meth == 2
        self.lineEditWavenumber.setEnabled(iswn)
        self.horizontalSliderWavenumber.setEnabled(iswn)
        wn = self.plot_raw.getWavenumbers()
        if wn is not None:
            wix = self.horizontalSliderWavenumber.value()
            if wn[0] > wn[-1]:
                wix = len(wn)-1 - wix
            self.plot_raw.setProjection(meth, wix)

    def selectedSpectraUpdated(self, n):
        self.spinBoxSpectra.setValue(n)
        self.plot_spectra.setData(self.plot_raw.getWavenumbers(), None,
                                  self.plot_raw.getSelectedData())

    def selectSpectra(self):
        self.plot_raw.setSelectedCount(
                self.spinBoxSpectra.value(),
                self.checkBoxAutopick.isChecked())

    def setImageCount(self, n):
        if n:
            self.horizontalSliderImage.setMaximum(n-1)
        else:
            self.horizontalSliderImage.setValue(0)
            self.lineEditImageInfo.setText('')
        self.horizontalSliderImage.setEnabled(n > 0)
        self.lineEditImageInfo.setEnabled(n > 0)
        self.horizontalSliderImage.setHidden(n < 2)

    def saveParameters(self, p):
        "Copy from UI to some kind of parameters object"
        p.plotMethod = self.comboBoxMethod.currentIndex()
        p.plotColors = self.comboBoxCmaps.currentText()
        p.plotWavenum = self.lineEditWavenumber.value()
        p.spectraCount = self.spinBoxSpectra.value()
        p.spectraAuto = self.checkBoxAutopick.isChecked()

    def loadParameters(self, p):
        "Copy from some kind of parameters object to UI"
        self.comboBoxMethod.setCurrentIndex(p.plotMethod)
        self.comboBoxCmaps.setCurrentText(p.plotColors)
        self.lineEditWavenumber.setValue(p.plotWavenum)
        self.checkBoxAutopick.setChecked(p.spectraAuto)
        self.spinBoxSpectra.setValue(p.spectraCount)
        self.wavenumberEdit()

class ImageVisualizer():
    """
    Mixin class that adds hyperspectral image display to an OctavvsMainWindow.

    Things assumed to exist in self:
        imageProjection
        plot_raw
        data

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.spatialMode = False

        self.imageVisualizer.pushButtonWhitelight.clicked.connect(
            self.loadWhite)
        self.imageVisualizer.horizontalSliderImage.valueChanged.connect(
            self.selectImage)
        self.whiteLightNames = {}

    def updateFile(self, num):
        super().updateFile(num)

        if self.data.pixelxy is not None:
            self.spatialMode = True
            kd = KDTree(self.data.pixelxy)
            dists = kd.query(self.data.pixelxy, k=2)[0][..., 1]
            adist = np.percentile(dists, 20)
            ps = self.imageVisualizer.lineEditPixelSize.value()
            adist = float(f'{adist:.2g}')
            if ps > adist:
                self.imageVisualizer.lineEditPixelSize.setValue(adist)
                self.imageVisualizer.pixelSizeEdit()
        else:
            self.spatialMode = False

        self.imageVisualizer.plot_raw.setData(
            self.data.wavenumber, self.data.raw,
            self.data.pixelxy)
        self.imageVisualizer.plot_raw.setDimensions(self.data.wh)

    def updateWavenumberLabels(self):
        minmax = ["%.2f" % self.data.wmin, "%.2f" % self.data.wmax]
        self.imageVisualizer.labelMinwn.setText(minmax[1 - self.plot_ltr])
        self.imageVisualizer.labelMaxwn.setText(minmax[0 + self.plot_ltr])

    def updateWavenumberRange(self):
        super().updateWavenumberRange()

        self.updateWavenumberLabels()
        wmin = min(self.data.wmin, constants.WMIN)
        wmax = max(self.data.wmax, constants.WMAX)
        self.imageVisualizer.lineEditWavenumber.setRange(
            wmin, wmax, default=.5*(wmin+wmax))
        # Make sure the sliders are in sync with the boxes
        self.imageVisualizer.wavenumberEdit()

    def updatedFile(self):
        self.imageVisualizer.updateSpectralImage()
        self.updateWhiteLightImages()
        self.imageVisualizer.selectSpectra()

    def updateDimensions(self, wh):
        try:
            self.imageVisualizer.plot_raw.setDimensions(wh)
        except ValueError:
            pass
        self.imageVisualizer.updateSpectralImage()

    def updateUnitsAndOrder(self, units, ltr):
        super().updateUnitsAndOrder(units, ltr)
        self.imageVisualizer.labelUnits1.setText(self.unitsRichText())
        self.imageVisualizer.plot_spectra.setOrder(ltr)
        self.imageVisualizer.horizontalSliderWavenumber.setInvertedAppearance(
            not ltr)
        self.updateWavenumberLabels()


    def updateWhiteLightImages(self):
        imgcnt = len(self.data.images) if self.data.images else 0
        intern = self.spatialMode or imgcnt
        if intern:
            self.imageVisualizer.setImageCount(imgcnt)
            self.selectImage(self.imageVisualizer.horizontalSliderImage.value())
            self.imageVisualizer.plot_whitelight.setHidden(self.spatialMode)
        else:
            fname = self.whiteLightNames[self.data.curFile] if \
                self.data.curFile in self.whiteLightNames else \
                    os.path.splitext(self.data.curFile)[0]+'.jpg'
            self.imageVisualizer.plot_whitelight.load(filename=fname)
            self.imageVisualizer.plot_whitelight.setHidden(
                not self.imageVisualizer.plot_whitelight.is_loaded())
        self.imageVisualizer.layoutSpatial.setHidden(not self.spatialMode
                                                     and imgcnt < 2)
        self.imageVisualizer.pushButtonWhitelight.setHidden(intern)
        self.imageVisualizer.labelPixelSize.setHidden(not self.spatialMode)
        self.imageVisualizer.lineEditPixelSize.setHidden(not self.spatialMode)

    # White light image stuff
    def loadWhite(self):
        filename = self.getImageFileName(title="Load white light image",
                                         settingname='whitelightDir')
        if not filename:
            return
        self.imageVisualizer.plot_whitelight.load(filename=filename)
        ok = self.imageVisualizer.plot_whitelight.is_loaded()
        self.imageVisualizer.plot_whitelight.setHidden(not ok)
        if ok:
            self.whi4teLightNames[self.data.curFile] = filename

    def selectImage(self, num):
        if self.data.images and len(self.data.images):
            if self.spatialMode:
                self.imageVisualizer.plot_raw.setImage(self.data.images[num])
            else:
                self.imageVisualizer.plot_whitelight.load(
                    image=self.data.images[num])
            self.imageVisualizer.lineEditImageInfo.setText(
                self.data.images[num].name)
        else:
            if self.spatialMode:
                self.imageVisualizer.plot_raw.setImage(None)
            else:
                self.imageVisualizer.plot_whitelight.clear()
            self.imageVisualizer.lineEditImageInfo.setText('[No image]')





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 28 01:07:18 2019

@author: carl
"""

from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import numpy as np
from collections import Counter


class PixelWidget(FigureCanvas):
    def __init__(self, parent=None):
        self.fig = Figure()
        self.ax = self.fig.add_subplot(111)
        self.ax.set_axis_off()
        self.ax.set_position([0,0,1,1])
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
        self.mpl_connect('button_press_event', self.onclick)
        self.wh = (0, 0)


    def newRect(self, xy, color):
        return patches.Rectangle((xy[0]-.45, xy[1]-.45), .9, .9,
                                 fill=False, color=color, linewidth=1)

    def onclick(self, event):
        if event.inaxes != self.ax:
            return
        x, y = (int(round(event.xdata)), int(round(event.ydata)))
        if not (0 <= x < self.wh[0] and 0 <= y < self.wh[1]):
            return
        self.clickPoint(x, y)


class WhiteLightWidget(PixelWidget):
    clickedPoint = pyqtSignal(int, int)

    def __init__(self, parent=None):
        PixelWidget.__init__(self, parent)
        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        self.selected = {}  # Map from (x,y) to Patch
        self.img = None

    @pyqtSlot(int, int)
    def removePoint(self, x, y):
        self.selected.pop((x, y)).remove()
        self.draw_idle()

    @pyqtSlot(int, int, str)
    def addPoint(self, x, y, color):
        xy = (x, y)
        p = self.newRect(xy, color)
        self.ax.add_patch(p)
        self.selected[xy] = p
        self.draw_idle()

    def clickPoint(self, x, y):
        self.clickedPoint.emit(x, y)

    def load(self, filename):
        if self.img is not None:
            self.img.remove()
            self.img = None
        try:
            img = plt.imread(filename)
            self.wh = (64, 64)
            self.img = self.ax.imshow(img, extent=(-.5, self.wh[0] - .5,
                                                   self.wh[1] - .5, -.5))
        except FileNotFoundError:
            pass
        self.draw_idle()


class ProjectionWidget(PixelWidget):
    removedPoint = pyqtSignal(int, int)
    addedPoint = pyqtSignal(int, int, str)
    alteredCounts = pyqtSignal(Counter)

    def __init__(self, parent=None):
        PixelWidget.__init__(self, parent)

        self.wavenumber = None
        self.raw = None
        self.projection = np.random.rand(16,16)
        self.cmap = 'hot'
        self.selected = {}  # Map from (x,y) to (annotation, Patch)
        self.curannot = 0
        self.mainimg = self.ax.imshow(self.projection, self.cmap, zorder=0)
        self.drawannot = None
        self.mpl_connect('button_release_event', self.onrelease)
        self.mpl_connect('motion_notify_event', self.onmotion)

    def setData(self, wn, raw, wh):
        self.wavenumber = wn
        self.raw = raw
        self.wh = wh

    def setAnnotation(self, annot):
        """
        Set the annotation to use for subsequently clicked points
        """
        self.curannot = annot

    def getWavenumbers(self):
        return self.wavenumber

    def setProjection(self, method, wavenumix):
        if self.wavenumber is None:
            return
        if method == 0:
            data = np.trapz(self.raw, self.wavenumber, axis=1)
        elif method == 1:
            data = self.raw.max(axis=1)
        elif method == 2:
            data = self.raw[:, wavenumix]
        data = data.reshape(self.wh)

        if self.projection.shape != data.shape:
            self.clearSelected()
            self.ax.clear()
            self.ax.set_axis_off()
            self.mainimg = self.ax.imshow(data, self.cmap, zorder=0)
        elif self.mainimg is not None:
            self.mainimg.set_data(data)
            self.mainimg.autoscale()
        self.projection = data
        self.draw_idle()

    def getAnnotations(self):
        ann = np.zeros(self.wh, dtype=int)
        for key, val in self.selected.items():
            ann[key] = val[0] + 1
        return ann

    def setAnnotations(self, ann):
        self.clearSelected()
        xys = np.transpose(np.nonzero(ann))
        for xy in xys:
            self.addPoint(tuple(xy), ann[tuple(xy)] - 1)

    def removePoint(self, xy):
        self.selected.pop(xy)[1].remove()
        self.removedPoint.emit(*xy)
        self.updateCounts()
        self.draw_idle()

    def rectColor(self, annot):
        colors = 'bgmykcrw'
        return colors[annot % len(colors)]

    def addPoint(self, xy, annot = None):
        if annot is None:
            annot = self.curannot
        color = self.rectColor(annot)
        p = self.newRect(xy, color)
        self.ax.add_patch(p)
        self.selected[xy] = (annot, p)
        self.addedPoint.emit(*xy, color)
        self.updateCounts()
        self.draw_idle()

    def clearSelected(self):
        for xy in self.selected.keys():
            self.removedPoint.emit(*xy)
        for ap in self.selected.values():
            ap[1].remove()
        self.selected = {}
        self.updateCounts()
        self.draw_idle()

    def updateCounts(self):
        self.alteredCounts.emit(Counter([x[0] for x in self.selected.values()]))

    @pyqtSlot('QString')
    def setCmap(self, s):
        self.cmap = s
        if self.mainimg:
            self.mainimg.set_cmap(s)
        self.draw_idle()

    def clickOnePoint(self, x, y):
        if (x, y) in self.selected:
            self.removePoint((x, y))
        else:
            self.addPoint((x, y))

    def clickPoint(self, x, y):
        if (x, y) in self.selected:
            self.removePoint((x, y))
            self.drawannot = False
        else:
            self.addPoint((x, y))
            self.drawannot = True

    def onmotion(self, event):
        if self.drawannot is None:
            return
        if event.inaxes != self.ax:
            return
        x, y = (int(round(event.xdata)), int(round(event.ydata)))
        if not (0 <= x < self.wh[0] and 0 <= y < self.wh[1]):
            return
        if (x, y) in self.selected:
            if not self.drawannot:
                self.removePoint((x, y))
        elif self.drawannot:
            self.addPoint((x, y))

    def onrelease(self, event):
        self.drawannot = None




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb  4 02:56:27 2021

@author: carl
"""

import numpy as np
import itertools
from statsmodels.multivariate import pca as smpca
from scipy.optimize import minimize

def correct_mirage(wn, data, breaks=[(973,977), (1201.5,1205.5), (1449,1453)],
                   endpoints=[[4,4],[4,4],[4,4]], slopefactor=[1, 1, 1],
                   pca=True, pca_ncomp=[2, 2, 2],
                   soft_limits=True, sl_level=[5, 5, 5],
                   chipweight=[0, .3, .35, .35]):
    """
    Correct for step discontinuities in PTIR spectra caused by pixel-dependent
    imperfections in focusing and alignment differentially affecting the
    four quantum cascade laser (QCL) chip regions of the mIRage machines.

    Correctly measured spectra are assumed to approximately follow a straight
    line around each transition between regions. For each spectrum, the level
    at each transition point is estimated from a small number of points in
    the region to the left, and likewise in the region to the right, and
    the ratio between these levels define the relative rescaling factor at
    that transition.

    The entire QCL regions are multiplicatively rescaled in the spectrum,
    with the overall absorption level approximately preserved through
    weighted zero-normalization of the log rescaling factors.

    Values in the transition regions are interpolated anew from the
    surrounding points.

    'A first chip provides IR light from 770 cm-1 to 975.8 cm-1; a
    second chip from 975.9 cm-1 to 1204 cm-1; a third chip from
    1204.1 cm-1 to 1451.2 cm-1 and a fourth chip from 1451.3
    cm-1 to 1802 cm-1.'

    Parameters
    ----------
    wn : array
        wavenumbers, sorted
    data : array(pixel, wn)
        raw mIRage data
    breaks : list of pairs, optional
        regions where the spectra are interpolated between the QCL chips
    endpoints: int or array-like, optional
        number of wavenumbers to consider when normalizing levels at
        the breaks between regions; can be given as a single int, a triplet
        or a triplet of pairs (for the left and right side of the break).
    slopefactor : float or array-like, optional
        the extent to which the slope of points near a break are used when
        inferring the level at the break point; useful range is 0 to 1.
    pca : bool
        perform PCA filtering across all spectra of the points in each
        transition region.
    pca_ncomp : int or array-like, optional
        number of components for PCA filter of intensity levels around
        each break; useful for reducing noise in the estimation.
    soft_limits : bool, optional
        limit the range of correction factors based on their estimated
        distribution and the reliability of individual spectra.
    sl_level : float or array-like, optional
        base level of the soft limits (in stddevs).
    chipweight : list of floats, optional
        weights of the chip regions when normalizing correction levels.

    Returns
    -------
    data : array
        corrected data
    shifts : array(npixels, nbreaks)
        log rescaling factor for the spectra at each of the breaks.

    """
    data = data.copy()
    if not len(data):
        return data, None
    flipwn = wn[0] > wn[-1]
    if flipwn:
        data = data[:, ::-1]
        wn = wn[::-1]

    def linreg(x, y):
        xm = x.mean()
        ym = y.mean(1)
        rx = x - xm
        s = (rx * (y.T - ym).T).sum(1) / (rx * rx).sum()
        return xm, ym, s

    cuts = [0]
    logshifts = []
    weights = []
    # endp = (np.ones((len(breaks), 2), dtype=int).T * np.transpose(endpoints)).T
    endp = np.broadcast_to(np.transpose(endpoints), (2, len(breaks))).T
    slopef = np.broadcast_to(np.transpose(slopefactor), (2, len(breaks))).T
    if pca:
        ncomp = np.broadcast_to(pca_ncomp, len(breaks))
    for j, bw in enumerate(breaks):
        # First index (in, after) the break region
        b = np.searchsorted(wn, bw[0]), np.searchsorted(wn, bw[1])
        cuts.extend(b)
        c = [max(0, b[0] - endp[j, 0]), b[0],
             b[1], min(b[1] + endp[j, 1], len(wn))]
        # Possibly do PCA noise reduction before computing slope
        dat = np.maximum(
            0, data[:, list(range(c[0], c[1])) + list(range(c[2], c[3]))])
        if pca and ncomp[j] and ncomp[j] < len(dat):
            dat = smpca.PCA(dat, ncomp=ncomp[j]).projection
            dat = np.maximum(0, dat)

        dleft = dat[:, :(c[1] - c[0])]
        dright = dat[:, (c[1] - c[0]):]
        wb = linreg(wn[c[0]:c[1]], dleft)
        we = linreg(wn[c[2]:c[3]], dright)
        bmid = (bw[0] + bw[1]) / 2
        # Compute desired step with some provision for near-zero values
        epsilon = 1e-3
        lb = wb[1] + slopef[j, 0] * (bmid - wb[0]) * wb[2]
        le = we[1] - slopef[j, 1] * (we[0] - bmid) * we[2]
        logshifts.append(np.log(np.maximum(epsilon * lb.mean(), lb)) -
                         np.log(np.maximum(epsilon * le.mean(), le)))

        weps = .1
        weights.append(np.minimum(
            dleft.mean(1) / (np.std(dleft, 1) + weps * dleft.mean()),
            dright.mean(1) / (np.std(dright, 1) + weps * dright.mean())))

    cuts.append(len(wn))
    cuts = np.array(cuts).reshape((-1, 2))
    logshifts = np.array(logshifts)

    means = None
    if soft_limits:
        sl_level = np.broadcast_to(sl_level, len(breaks))
        sl_level = np.maximum(sl_level, 0)
        # Weight/importance: values above a fraction of the mean
        # weights = np.array(weights)
        weights = np.array(weights)
        # Compute weighted mean and stddev
        wsum = weights.sum(1)
        means = ((logshifts * weights).sum(1) / wsum)[:, None]
        sd = (((logshifts - means) ** 2 * weights).sum(1) / wsum) ** .5
        # or
        # sd = (((logshifts - means) ** 2 * weights**2).sum(1) / wsum**2) ** .5

        lim = (sd * sl_level)[:, None]
        logshifts = logshifts - means
        logshifts = means + logshifts / (1 + (logshifts / lim)**2)**.5

    # Turn steps into scale factors for the regions
    chipscale = np.zeros((len(breaks) + 1, len(data)))
    for j, ls in enumerate(logshifts):
        chipscale[j+1, :] = chipscale[j, :] + logshifts[j]
    chipscale = np.exp(chipscale)
    # Normalize levels to preserve weighted mean level
    chipscale = chipscale / (chipweight @ chipscale)

    # Rescale data and interpolate in gaps
    pc = None
    for j, c in enumerate(cuts):
        data[:, c[0]:c[1]] *= chipscale[j, :, None]
        if j:
            data[:, pc:c[0]] = (np.linspace(data[:, pc-1], data[:, c[0]],
                                           num=c[0]-pc+2)[1:-1]).T
        pc = c[1]

    if flipwn:
        data = data[:,::-1]
    return np.maximum(0, data), (logshifts, weights)


def optimize_mirage_pca(wn, data, callback=None, **cm_args):
    def mirage_variance():
        corr = correct_mirage(wn, data, **cm_args)[0]
        normsd = np.std(corr / corr.mean(1)[:, None], 0).mean()
        return normsd

    nb = len(cm_args['breaks'])
    pca_ncomps = itertools.product(*([range(0, 4)] * nb))
    bestsol = None
    cm_args['pca'] = True
    for i, pca_ncomp in enumerate(pca_ncomps):
        cm_args['pca_ncomp'] = pca_ncomp
        v = mirage_variance()
        if bestsol is None or v < bestsol[0]:
            bestsol = (v, pca_ncomp)
        if callback is not None:
            callback(i)
    return bestsol

def optimize_mirage_sl(wn, data, callback=None, **cm_args):
    nb = len(cm_args['breaks'])
    lims = (.001, 10)

    def mirage_variance(x):
        x = np.maximum(lims[0], np.minimum(lims[1], x))
        cm_args['sl_level'] = x
        corr = correct_mirage(wn, data, **cm_args)[0]
        normsd = np.std(corr / corr.mean(1)[:, None], 0).mean()
        return normsd

    init = np.ones((nb + 1, nb)) * 2.
    init[1:, :] -= np.eye(nb)
    cm_args['soft_limits'] = True
    optr = minimize(mirage_variance, method='Nelder-Mead', x0=init[0],
                    tol=1e-2, callback=callback,
                    options={'xatol': .05, 'initial_simplex': init})
    if not optr.success:
        raise RuntimeError('Optimization failed: ' + optr.message)
    return optr.fun, np.maximum(lims[0], np.minimum(lims[1], optr.x))

def optimize_mirage_endpoints(wn, data, callback=None, **cm_args):
    nb = len(cm_args['breaks'])
    lims = (4, 17)

    def mirage_variance(x):
        cm_args['endpoints'] = np.maximum(lims[0], np.minimum(
            lims[1], x, dtype=int, casting='unsafe'))
        corr = correct_mirage(wn, data, **cm_args)[0]
        normsd = np.std(corr / corr.mean(1)[:, None], 0).mean()
        return normsd

    init = np.ones((nb + 1, nb)) * 9.
    init[1:, :] -= np.eye(nb) * 3
    optr = minimize(mirage_variance, method='Nelder-Mead', x0=init[0],
                    tol=1e-4, callback=callback,
                    bounds=[lims] * nb,
                    options={'xatol': 1, 'initial_simplex': init})
    if not optr.success:
        raise RuntimeError('Optimization failed: ' + optr.message)
    finalx = np.maximum(lims[0], np.minimum(
        lims[1], optr.x, dtype=int, casting='unsafe'))
    return optr.fun, finalx

